""" Module for log controllers """
from datetime import datetime
from typing import List
from flask import redirect, render_template, request, url_for
from databus.client.client import Client
from databus.database.abstract_database import LogListItem
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError


class ClientLog: # pylint: disable=R0903
    """ Defines a client and all log files present """
    def __init__(self, p_client: Client = None, p_logs: List[LogListItem] = None):
        if p_client is None:
            self.client = Client()
        else:
            self.client = p_client

        if p_logs is None:
            self.logs = []
        else:
            self.logs = p_logs


class ClientLogReader:
    """ Class dealing with client log files """
    def __init__(self, p_dispatcher: AbstractDispatcher):
        self._dispatcher = p_dispatcher

    def get_client_log_content(self, p_client_id: str, p_log_id: str) -> str:
        """ Returns content of the log file """
        client_db = self._dispatcher.get_client_database(p_client_id)
        log_content = client_db.get_log_content(p_log_id)
        return log_content

    def get_client_log_list(self, p_client_id: str = None) -> List[ClientLog]:
        """ Returns a list of client - log files """
        output = []

        for client in self._dispatcher.all_clients:
            if p_client_id is not None and client.id != p_client_id:
                continue
            client_database = self._dispatcher.get_client_database(client.id)
            logs = client_database.get_log_list()
            logs.sort(key=lambda r: r.log_id, reverse=True)
            output.append(ClientLog(client, logs))
        return output


class LogDisplayController(AbstractController):
    """ Log display """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        log_file = request.args.get("log", 0, type=str)
        file_content = ClientLogReader(self.dispatcher).get_client_log_content(self.requested_client_id, log_file).replace("\n", "<br><br>") # pylint: disable=C0301
        return file_content


class LogListController(AbstractController):
    """ Log list """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_root()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        log_reader = ClientLogReader(self.dispatcher)
        entries = log_reader.get_client_log_list(p_client_id=self.authenticated_client_id)

        return render_template(
            "log_list.html",
            entries=entries,
            alias=self.dispatcher.ticket.system_alias)


class LogPurgeController(AbstractController):
    """ Log purge """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        self.requested_client_database.delete_old_logs(datetime.now())
        return redirect(url_for("_log_list"), code=302)



""" Excel attachment filter module """
from typing import List
from databus.passenger.email import Email
from databus.pqueue.queue_status import PassengerQueueStatus, QueueStatus
from databus.processor.abstract_processor import AbstractProcessor


class EmailFilterExcelAttachment(AbstractProcessor): # pylint: disable=R0903
    """ Excel attachment filter class """

    def process(self, p_passengers: List[PassengerQueueStatus]):
        """ Marks E-Mails without Excel attachments as complete """
        self.log.append_text("Filtering E-Mails by Excel attachment")

        for pqs in p_passengers:
            email = pqs.passenger
            if EmailFilterExcelAttachment._has_excel_attachment(email):
                self.log.append_text(email.id_text + " has an Excel attachment")
                pqs.set_processor_status(self.__module__, QueueStatus.complete)
            else:
                self.log.append_text((email.id_text + " doesn't have Excel attachments"))
                pqs.set_all_processor_statuses(QueueStatus.complete)
                pqs.set_all_pusher_statuses(QueueStatus.complete)

    @staticmethod
    def _has_excel_attachment(received_email: Email) -> bool:
        excel_attachments = received_email.excel_attachments
        if len(excel_attachments) <= 0:
            return False
        return True



""" Module to build update conditions """
from databus.database.sql_db.sql_database_arguments import SqlDatabaseArguments
from databus.database.sql_db.modifiable import Modifiable
from databus.database.sql_db.where_builder import WhereBuilder

class UpdateBuilder(Modifiable):
    """ Helper class to build update commands """
    def __init__(self, p_args: SqlDatabaseArguments, p_client_id: str = ""):
        super().__init__(p_args, p_client_id)
        self.where = WhereBuilder(p_client_id)

    @property
    def update_command(self) -> str:
        """ Update command """
        query = "UPDATE " + self.table + " SET "

        first_val = True
        for keyval in self._key_values:
            if not first_val:
                query += ", "
            query += keyval.key + " = "
            if keyval.is_string:
                query += "'"
            query += Modifiable._get_safe_string(keyval.val)
            if keyval.is_string:
                query += "' "
            first_val = False

        query += self.where.where
        return query


    def clear(self):
        """ Reset """
        super().clear()
        try:
            self.where.clear()
        except Exception: # pylint: disable=W0703
            pass



""" Module to check differences between memory and database states """
from enum import Enum
from typing import List


class Action(Enum):
    """ Defines a database action """
    INSERT = 1
    UPDATE = 2
    DELETE = 3


class AnalysisResult(Enum):
    """ Analysis result """
    NOT_FOUND = 1
    MODIFIED = 2
    SAME = 3


class DifferenceResult: # pylint: disable=R0903
    """ Result of a difference check """
    def __init__(self, p_table: str, p_row: dict, p_action: Action):
        self.table = p_table
        self.row = p_row
        self.action = p_action


class TableKey: # pylint: disable=R0903
    """ A list of table keys """
    def __init__(self, p_table: str, p_keys: List[str]):
        self.table = p_table
        self.keys = p_keys


class DifferenceChecker: # pylint: disable=R0903
    """ Class to check differences between memory and database states
    This class will take 2 dictionaries.
    A) Dictionary in memory
    B) Dictionary in database
    It will compare both, and tell what needs to be inserted / updated / deleted
    """

    def __init__(self, p_keys: List[TableKey], p_memory: dict, p_database: dict):
        self._result = []
        self._table_keys = p_keys
        self._memory = p_memory
        self._database = p_database
        self._compare_database_with_memory()
        self._compare_memory_with_database()

    @property
    def result(self) -> List[DifferenceResult]:
        """ Returns the result of the class """
        return self._result

    def _analyse(self, p_table_name: str, p_src_line: dict, p_tar: dict) -> AnalysisResult:
        if p_table_name not in p_tar:
            return AnalysisResult.NOT_FOUND
        tar_lines = p_tar[p_table_name]

        table_key = None
        for table_key_candidate in self._table_keys:
            if table_key_candidate.table == p_table_name:
                table_key = table_key_candidate
                break

        if table_key is None:
            raise Exception("No key provided for table " + p_table_name)

        target_found = False
        for tar_line in tar_lines:
            this_is_target_line = True
            for key_field in table_key.keys:
                if p_src_line[key_field] != tar_line[key_field]:
                    this_is_target_line = False
                    break
            if not this_is_target_line:
                continue
            target_found = True

            for column in tar_line:
                if p_src_line[column] != tar_line[column]:
                    return AnalysisResult.MODIFIED

        if target_found:
            return AnalysisResult.SAME
        return AnalysisResult.NOT_FOUND

    def _compare_memory_with_database(self):
        for table_name in self._memory:
            for table_row in self._memory[table_name]:
                result = self._analyse(
                    table_name,
                    table_row,
                    self._database)

                if result == AnalysisResult.MODIFIED:
                    diff_result = DifferenceResult(table_name, table_row, Action.UPDATE)
                    self._result.append(diff_result)
                if result == AnalysisResult.NOT_FOUND:
                    diff_result = DifferenceResult(table_name, table_row, Action.INSERT)
                    self._result.append(diff_result)

    def _compare_database_with_memory(self):
        for table_name in self._database:
            for table_row in self._database[table_name]:
                result = self._analyse(
                    table_name,
                    table_row,
                    self._memory)

                if result == AnalysisResult.NOT_FOUND:
                    diff_result = DifferenceResult(table_name, table_row, Action.DELETE)
                    self._result.append(diff_result)



""" Abstract factory module """
from abc import ABC, abstractmethod
from enum import Enum
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher, DispatcherTicket



class DispatcherCreationError(Exception):
    """ Dispatcher creation exception """

    class ErrorCode(Enum):
        """ Error code """
        cant_create_instance: 1
        parameter_missing: 2

    def __init__(self, p_error_code: ErrorCode, p_module: str = None):
        super().__init__()
        self.error_code = p_error_code

        if p_module is None:
            self.module = ""
        else:
            self.module = p_module

    @property
    def message(self) -> str:
        """ Error message """
        if self.error_code == DispatcherCreationError.ErrorCode.cant_create_instance:
            return "Can't create " + self.module + " dispatcher instance"
        return "Dispatcher creation error"


class AbstractDispatcherFactory(ABC): # pylint: disable=R0903
    """ Abstract dispatcher factory class """
    @abstractmethod
    def create_dispatcher(self, p_module: str, p_ticket: DispatcherTicket) -> AbstractDispatcher:
        """ Creates a new dispatcher """



""" Client passenger module """
from datetime import datetime, timedelta
from typing import List


class ClientPassenger:  # pylint: disable=R0903
    """ Client passenger class """
    def __init__(self, # pylint: disable=R0913
                 p_name: str = "Undefined",
                 p_puller_modules: List[str] = None,
                 p_queue_module: str = None,
                 p_processor_modules: List[str] = None,
                 p_pusher_modules: List[str] = None,
                 p_sync_frequency: int = 0,
                 p_queue_life_span: int = 0):

        self.name = p_name
        self.sync_frequency = p_sync_frequency
        self.queue_life_span = p_queue_life_span

        if p_queue_module is None:
            self.queue_module = ""
        else:
            self.queue_module = p_queue_module

        if p_puller_modules is None:
            self.puller_modules = []
        else:
            self.puller_modules = p_puller_modules

        if p_processor_modules is None:
            self.processor_modules = []
        else:
            self.processor_modules = p_processor_modules

        if p_pusher_modules is None:
            self.pusher_modules = []
        else:
            self.pusher_modules = p_pusher_modules

    @property
    def queue_expiry_date(self) -> datetime:
        """ Returns queue expiry date """
        output = datetime.now() - timedelta(days=self.queue_life_span)
        return output



""" Queue module for JSON database """
from datetime import datetime
from enum import Enum
import json
from os import mkdir, path, remove
import shutil
from typing import List
import uuid
from databus.client.log import Log, LogEntry, MessageType
from databus.database.json_db.json_database_arguments import JsonDatabaseArguments
from databus.database.json_db.json_path_builder import JsonPathBuilder
from databus.database.json_db.json_toolkit import JsonToolkit
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.passenger.attachment import Attachment, AttachmentError, AttachmentFormat
from databus.passenger.attachment import Validator as AttachmentValidator
from databus.passenger.abstract_factory import AbstractPassengerFactory
from databus.pqueue.queue_status import \
    QueueStatus, PassengerQueueStatus, ProcessorQueueStatus, PusherQueueStatus
from databus.pqueue.queue_status import Validator as QueueStatusValidator



class PassengerError(Exception):
    """ Base exception class for passenger errors """
    class ErrorCode(Enum):
        """ Error code enum """
        internal_id_missing = 1
        puller_module_missing = 2
        already_exists = 3

    def __init__(self, p_error_code: ErrorCode, p_passenger_id: str = None):
        super().__init__()
        self.error_code = p_error_code

        if p_passenger_id is None:
            self.passenger_id = ""
        else:
            self.passenger_id = p_passenger_id

    @property
    def message(self) -> str:
        """ Returns error message as text """
        if self.error_code == PassengerError.ErrorCode.internal_id_missing:
            return "Passenger " + self.passenger_id + " is missing internal ID"
        if self.error_code == PassengerError.ErrorCode.puller_module_missing:
            return "Passenger " + self.passenger_id + " is missing puller module"
        if self.error_code == PassengerError.ErrorCode.already_exists:
            return "Passenger " + self.passenger_id + " exists already"
        return "Passenger error"


class JsonQueue:
    """ Queue implementation for JSON database """
    class DataOperation(Enum):
        """ Data operation enum """
        insert = 1
        update = 2

    def __init__(self,
                 p_client_id: str,
                 p_log: Log,
                 p_passenger_factory: AbstractPassengerFactory,
                 p_args: JsonDatabaseArguments):
        self.client_id = p_client_id
        self._log = p_log
        self._passenger_factory = p_passenger_factory
        self._args = p_args
        self._path = JsonPathBuilder(self.client_id, self._args)

    def delete_passengers(self, p_passengers: List[AbstractPassenger]):
        """ Deletes the given passengers from the disk """
        all_passenger_directories = self._path.passenger_directories
        for passenger in p_passengers:
            if passenger.internal_id in all_passenger_directories:
                self._log.append_text("Deleting passenger directory " + passenger.internal_id)
                passenger_dir_path = self._path.get_passenger_directory_path(passenger.internal_id)
                shutil.rmtree(passenger_dir_path)

    def erase_passenger_queue(self):
        """ Deletes all passengers from the disk """
        all_passenger_directories = self._path.passenger_directories
        for passenger_directory in all_passenger_directories:
            self._log.append_text("Deleting passenger directory " + passenger_directory)
            passenger_dir_path = self._path.get_passenger_directory_path(passenger_directory)
            shutil.rmtree(passenger_dir_path)

    def get_passengers(self, # pylint: disable=R0912, R0913, R0914
                       p_passenger_module: str = None,
                       p_processor_status: QueueStatus = None,
                       p_pusher_status: QueueStatus = None,
                       p_puller_notified: bool = None,
                       p_pulled_before: datetime = None,
                       p_internal_id: str = None
                       ) -> List[PassengerQueueStatus]:
        """ Returns passengers """
        output = []

        for passenger_directory in self._path.passenger_directories:
            if p_internal_id is not None and passenger_directory != p_internal_id:
                continue

            passenger_json = self._get_passenger_file_as_json(passenger_directory)
            pull_datetime = JsonToolkit.convert_json_date_to_datetime(passenger_json["pull_datetime"]) # pylint: disable=C0301

            if p_pusher_status is not None and len(passenger_json["pusher_statuses"]) > 0:
                a_pusher_found = False
                for pusher_status in passenger_json["pusher_statuses"]:
                    if pusher_status["status"] == p_pusher_status.name:
                        a_pusher_found = True
                        break
                if not a_pusher_found:
                    continue

            if p_processor_status is not None and len(passenger_json["processor_statuses"]) > 0:
                a_processor_found = False
                for processor_status in passenger_json["processor_statuses"]:
                    if processor_status["status"] == p_processor_status.name:
                        a_processor_found = True
                        break
                if not a_processor_found:
                    continue

            if p_passenger_module is not None and passenger_json["passenger_module"] != p_passenger_module: # pylint: disable=C0301
                continue

            if p_puller_notified is not None and passenger_json["puller_notified"] != p_puller_notified: # pylint: disable=C0301
                continue

            if p_pulled_before is not None and pull_datetime > p_pulled_before:
                continue

            passenger_obj = self._passenger_factory.create_passenger(passenger_json["passenger_module"]) # pylint: disable=C0301
            passenger_obj.internal_id = passenger_json["internal_id"]
            passenger_obj.external_id = passenger_json["external_id"]
            passenger_obj.source_system = passenger_json["source_system"]
            passenger_obj.puller_module = passenger_json["puller_module"]
            passenger_obj.pull_datetime = pull_datetime
            self._log.append_text("Found passenger " + passenger_obj.id_text)

            for attachment_json in passenger_json["attachments"]:
                attachment_obj = self._get_attachment_obj(passenger_obj.internal_id,
                                                          attachment_json)
                passenger_obj.attachments.append(attachment_obj)

            if "log_guids" in passenger_json:
                for log_guid in passenger_json["log_guids"]:
                    passenger_obj.collect_log_guid(uuid.UUID(log_guid))

            paqs = PassengerQueueStatus(p_passenger=passenger_obj,
                                        p_pusher_statuses=[],
                                        p_processor_statuses=[])

            paqs.puller_notified = passenger_json["puller_notified"]

            for processor_status in passenger_json["processor_statuses"]:
                paqs.processor_statuses.append(ProcessorQueueStatus(
                    p_processor_module=processor_status["processor_module"],
                    p_status=QueueStatus[processor_status["status"]]))

            for pusher_status in passenger_json["pusher_statuses"]:
                paqs.pusher_statuses.append(PusherQueueStatus(
                    p_pusher_module=pusher_status["pusher_module"],
                    p_status=QueueStatus[pusher_status["status"]]))

            output.append(paqs)
        return output

    def insert_passenger(self, p_passenger_status: PassengerQueueStatus):
        """ Creates a new folder & puts files within """
        self._log.append_text("Adding passenger " + p_passenger_status.passenger.id_text + " to queue") # pylint: disable=C0301
        self._validate_passenger_status(p_passenger_status, JsonQueue.DataOperation.insert)

        passenger_dict = {
            "external_id": p_passenger_status.passenger.external_id,
            "internal_id": str(p_passenger_status.passenger.internal_id),
            "source_system": p_passenger_status.passenger.source_system,
            "passenger_module": p_passenger_status.passenger.passenger_module,
            "puller_module": p_passenger_status.passenger.puller_module,
            "puller_notified": p_passenger_status.puller_notified,
            "pull_datetime": p_passenger_status.passenger.pull_datetime.isoformat(),
            "attachments": [],
            "log_guids": [],
            "processor_statuses": [],
            "pusher_statuses": []
        }

        mkdir(self._path.get_passenger_directory_path(passenger_dict["internal_id"]))
        mkdir(self._path.get_attachment_directory_path(passenger_dict["internal_id"]))

        for attachment in p_passenger_status.passenger.attachments:
            attachment_dict = {
                "format": attachment.format.name,
                "name": attachment.name
            }
            passenger_dict["attachments"].append(attachment_dict)

            if attachment.format == AttachmentFormat.binary:
                self._write_attachment_file_bin(attachment_dict["name"],
                                                attachment.binary_content,
                                                passenger_dict["internal_id"])
            elif attachment.format == AttachmentFormat.text:
                self._write_attachment_file_text(attachment_dict["name"],
                                                 attachment.text_content,
                                                 passenger_dict["internal_id"])
            else:
                raise AttachmentError(AttachmentError.ErrorCode.invalid_format, attachment.format)

        for guid in p_passenger_status.passenger.log_guids:
            passenger_dict["log_guids"].append(str(guid))

        for processor_status in p_passenger_status.processor_statuses:
            processor_status_dict = {
                "processor_module": processor_status.processor_module,
                "status": processor_status.status.name
            }
            passenger_dict["processor_statuses"].append(processor_status_dict)

        for pusher_status in p_passenger_status.pusher_statuses:
            pusher_status_dict = {
                "pusher_module": pusher_status.pusher_module,
                "status": pusher_status.status.name
            }
            passenger_dict["pusher_statuses"].append(pusher_status_dict)

        self._write_passenger_json_into_file(passenger_dict)

    def set_processor_status(self,
                             p_passenger: AbstractPassenger,
                             p_processor_module: str,
                             p_status: QueueStatus):
        """ Updates processor status within the JSON file """

        self._log.append_text("Setting status " +
                              p_status.name +
                              " for passenger " +
                              p_passenger.id_text +
                              " processor " +
                              p_processor_module)

        passenger_json = self._get_passenger_file_as_json(p_passenger.internal_id)

        for processor_status in passenger_json["processor_statuses"]:
            if processor_status["processor_module"] == p_processor_module:
                processor_status["status"] = p_status.name

        self._write_passenger_json_into_file(passenger_json)

    def set_pusher_status(self,
                          p_passenger: AbstractPassenger,
                          p_pusher_module: str,
                          p_status: QueueStatus):
        """ Updates pusher status within the JSON file """
        self._log.append_text("Setting status " +
                              p_status.name +
                              " for passenger " +
                              p_passenger.id_text +
                              " pusher " +
                              p_pusher_module)

        passenger_json = self._get_passenger_file_as_json(p_passenger.internal_id)

        for pusher_status in passenger_json["pusher_statuses"]:
            if pusher_status["pusher_module"] == p_pusher_module:
                pusher_status["status"] = p_status.name

        self._write_passenger_json_into_file(passenger_json)

    def update_passenger(self, p_passenger_status: PassengerQueueStatus):
        """ Updates passenger """
        self._log.append_text("Updating passenger " + p_passenger_status.passenger.id_text)
        self._validate_passenger_status(p_passenger_status, JsonQueue.DataOperation.update)
        self.delete_passengers([p_passenger_status.passenger])
        self.insert_passenger(p_passenger_status)

    def _delete_attachment_file(self, p_file_name: str, p_internal_id: str):
        self._log.append_text("Deleting attachment " + p_file_name)

        full_path = self._path.get_attachment_file_path(p_internal_id, p_file_name)
        if not path.exists(full_path):
            self._log.append_entry(LogEntry(p_message="File not found", p_type=MessageType.warning))
            return
        remove(full_path)

    def _get_attachment_obj(self, p_internal_id: str, p_attachment_json: {}) -> Attachment:
        output = Attachment(p_name=p_attachment_json["name"],
                            p_format=AttachmentFormat[p_attachment_json["format"]])

        full_path = self._path.get_attachment_file_path(p_internal_id, output.name)
        self._log.append_text("Reading attachment from disk: " + full_path)

        if output.format == AttachmentFormat.text:
            with open(full_path, "r") as text_file:
                output.text_content = text_file.read()
        elif output.format == AttachmentFormat.binary:
            with open(full_path, "rb") as bin_file:
                output.binary_content = bin_file.read()
        else:
            raise AttachmentError(AttachmentError.ErrorCode.invalid_format, output.format)

        return output

    def _get_passenger_file_as_json(self, p_internal_id: str) -> dict:
        try:
            passenger_file_path = self._path.get_passenger_file_path(p_internal_id)
            self._log.append_text("Reading passenger file " + passenger_file_path)
            with open(passenger_file_path) as json_file:
                passengers_json = json.load(json_file)
            return passengers_json
        except Exception as error: # pylint: disable=W0703
            print(str(error))
            return {}

    def _validate_passenger_status(self,
                                   p_passenger_status: PassengerQueueStatus,
                                   p_operation: DataOperation):
        self._log.append_text("Validating passenger status for " + p_passenger_status.passenger.id_text) # pylint: disable=C0301

        if str(p_passenger_status.passenger.internal_id) == "":
            raise PassengerError(PassengerError.ErrorCode.internal_id_missing,
                                 p_passenger_id=p_passenger_status.passenger.id_text)

        QueueStatusValidator.validate_queue_module(
            "Puller",
            p_passenger_status.passenger.id_text,
            p_passenger_status.passenger.puller_module)

        for processor_status in p_passenger_status.processor_statuses:
            QueueStatusValidator.validate_queue_module(
                "Processor",
                p_passenger_status.passenger.id_text,
                processor_status.processor_module)

        for pusher_status in p_passenger_status.pusher_statuses:
            QueueStatusValidator.validate_queue_module(
                "Pusher",
                p_passenger_status.passenger.id_text,
                pusher_status.pusher_module)

        AttachmentValidator.validate_attachments(p_passenger_status.passenger.attachments)

        if p_operation == JsonQueue.DataOperation.insert and \
                p_passenger_status.passenger.internal_id in self._path.passenger_directories:
            raise PassengerError(PassengerError.ErrorCode.already_exists,
                                 p_passenger_status.passenger.id_text)

    def _write_attachment_file_text(self,
                                    p_file_name: str,
                                    p_file_content: str,
                                    p_internal_id: str):
        full_path = self._path.get_attachment_file_path(p_internal_id, p_file_name)
        self._log.append_text("Writing text attachment to disk: " + full_path)
        with open(full_path, "w") as text_file:
            text_file.write(p_file_content)

    def _write_attachment_file_bin(self,
                                   p_file_name: str,
                                   p_file_content: bytearray,
                                   p_internal_id: str):
        full_path = self._path.get_attachment_file_path(p_internal_id, p_file_name)
        self._log.append_text("Writing binary attachment to disk: " + full_path)
        with open(full_path, "wb") as bin_file:
            bin_file.write(p_file_content)

    def _write_passenger_json_into_file(self, p_json: dict):
        passenger_file_path = self._path.get_passenger_file_path(p_json["internal_id"])
        self._log.append_text("Writing passenger file to disk: " + passenger_file_path)
        with open(passenger_file_path, "w") as json_file:
            json.dump(p_json, json_file)



""" Default driver module """
from databus.client.log import Log, LogEntry, MessageType
from databus.driver.abstract_driver import AbstractDriver, BusTicket
from databus.pqueue.abstract_queue import AbstractQueue
from databus.pqueue.abstract_factory import AbstractQueueFactory
from databus.pqueue.queue_status import QueueStatus
from databus.processor.abstract_factory import AbstractProcessorFactory
from databus.puller.abstract_factory import AbstractPullerFactory
from databus.pusher.abstract_factory import AbstractPusherFactory


class Bus: # pylint: disable=R0903
    """ Bus object """
    def __init__(self,
                 p_ticket: BusTicket = BusTicket(),
                 p_queue: AbstractQueue = None):
        self.ticket = p_ticket
        self.new_passengers = []
        self.deliverable_passengers = []
        self.processable_passengers = []
        self.queue = p_queue


class PrimalDriver(AbstractDriver):
    """ Default driver class """

    def __init__(self,
                 p_queue_factory: AbstractQueueFactory,
                 p_processor_factory: AbstractProcessorFactory,
                 p_puller_factory: AbstractPullerFactory,
                 p_pusher_factory: AbstractPusherFactory):
        super().__init__(p_queue_factory, p_processor_factory, p_puller_factory, p_pusher_factory)
        self._bus = Bus()

    def drive(self, p_bus_ticket: BusTicket):
        """ Pulls passengers from source systems and pushes them to target systems """
        p_bus_ticket.log.append_text("Driving...")
        self._bus.ticket = p_bus_ticket

        try:
            queue = self.queue_factory.create_queue(
                p_bus_ticket.client_passenger.queue_module,
                p_bus_ticket.database,
                p_bus_ticket.log)
        except Exception as queue_error: # pylint: disable=W0703
            self._log_exception(queue_error, p_log=p_bus_ticket.log)
            return

        self._bus = Bus(p_ticket=p_bus_ticket, p_queue=queue)
        self._delete_old_log_files()
        self._pull_new_passengers()
        self._seat_passengers()
        self._notify_pullers_about_seated_passengers()
        self._read_processable_passengers()
        self._process_passengers()
        self._read_deliverable_passengers()
        self._push_passengers()

        p_bus_ticket.log.append_text("Drive complete")

    @property
    def queue(self) -> AbstractQueue:
        """ Queue object """
        return self._bus.queue

    def _delete_old_log_files(self):
        try:
            self._bus.ticket.log.append_text("Deleting old log files")
            log_expiry_date = self._bus.ticket.database.client.log_expiry_date
            self._bus.ticket.database.delete_old_logs(log_expiry_date)
        except Exception as del_error: # pylint: disable=W0703
            self._log_exception(del_error)

    def _log_exception(self, p_exception: Exception, p_log: Log = None):
        if p_log is None:
            log = self._bus.ticket.log
        else:
            log = p_log

        log.append_entry(LogEntry(p_message="Driver error: " + str(p_exception),
                                  p_type=MessageType.error))

    def _notify_pullers_about_seated_passengers(self):
        self._bus.ticket.log.append_text("Notifying pullers about seated passengers")
        try:
            for npass in self._bus.queue.get_puller_notifiable_passengers():
                try:
                    self._bus.ticket.log.append_text("Notifying " +
                                                     npass.passenger.puller_module +
                                                     " about " +
                                                     npass.passenger.id_text)

                    puller_obj = self.puller_factory.create_puller(
                        npass.passenger.puller_module,
                        self._bus.ticket.log)

                    puller_obj.notify_passengers_seated([npass.passenger])
                    npass.puller_notified = True
                    self._bus.queue.update_passenger_status(npass)
                except Exception as notif_error: # pylint: disable=W0703
                    self._log_exception(notif_error)
        except Exception as notif_error: # pylint: disable=W0703
            self._log_exception(notif_error)

    def _process_passengers(self):
        try:
            self._bus.ticket.log.append_text("Processing unprocessed passengers")

            for processable_passenger in self._bus.processable_passengers:
                self._bus.ticket.log.append_text(
                    "Processing " + processable_passenger.passenger.id_text)

                for processor_status in processable_passenger.processor_statuses:
                    if processor_status.status == QueueStatus.complete:
                        continue
                    self._bus.ticket.log.append_text("Processing " +
                                                     processable_passenger.passenger.id_text +
                                                     " via " +
                                                     processor_status.processor_module)
                    try:
                        processor_obj = self.processor_factory.create_processor(
                            processor_status.processor_module,
                            self._bus.ticket.log)
                        processor_obj.process([processable_passenger])
                        processor_status.status = QueueStatus.complete
                        self._bus.queue.update_passenger_status(processable_passenger)
                    except Exception as process_error: # pylint: disable=W0703
                        self._log_exception(process_error)
        except Exception as process_error: # pylint: disable=W0703
            self._log_exception(process_error)

    def _pull_new_passengers(self):
        try:
            self._bus.ticket.log.append_text("Pulling new passengers")
            for puller_module in self._bus.ticket.client_passenger.puller_modules:
                self._bus.ticket.log.append_text("Pulling via " + puller_module)

                new_passengers = self.pull_passengers_from_module(
                    puller_module,
                    p_log=self._bus.ticket.log)

                for new_passenger in new_passengers:
                    self._bus.new_passengers.append(new_passenger)
                    self._bus.ticket.log.append_text("Got new passenger: " + new_passenger.id_text)
        except Exception as pull_error: # pylint: disable=W0703
            self._log_exception(pull_error)

    def _push_passengers(self):
        try:
            self._bus.ticket.log.append_text("Pushing deliverable passengers")

            for deliverable_passenger in self._bus.deliverable_passengers:
                self._bus.ticket.log.append_text(
                    "Delivering " + deliverable_passenger.passenger.id_text)

                for pusher_status in deliverable_passenger.pusher_statuses:
                    if pusher_status.status == QueueStatus.complete:
                        continue
                    self._bus.ticket.log.append_text("Delivering " +
                                                     deliverable_passenger.passenger.id_text +
                                                     " via " +
                                                     pusher_status.pusher_module)
                    try:
                        pusher_obj = self.pusher_factory.create_pusher(pusher_status.pusher_module,
                                                                       self._bus.ticket.log)
                        pusher_obj.push(deliverable_passenger)
                        pusher_status.status = QueueStatus.complete
                        self._bus.queue.update_passenger_status(deliverable_passenger)
                    except Exception as push_error: # pylint: disable=W0703
                        self._log_exception(push_error)
        except Exception as push_error: #pylint: disable=W0703
            self._log_exception(push_error)

    def _read_deliverable_passengers(self):
        try:
            self._bus.ticket.log.append_text("Reading deliverable passengers from queue")

            self._bus.deliverable_passengers = self._bus.queue.get_deliverable_passengers(
                self._bus.ticket.client_passenger.name)

            for undelivered_passenger in self._bus.deliverable_passengers:
                self._bus.ticket.log.append_text(
                    "Undelivered passenger: " + undelivered_passenger.passenger.id_text)
        except Exception as read_error: # pylint: disable=W0703
            self._log_exception(read_error)

    def _read_processable_passengers(self):
        try:
            self._bus.ticket.log.append_text("Reading processable passengers from queue")

            self._bus.processable_passengers = self._bus.queue.get_processable_passengers(
                self._bus.ticket.client_passenger.name)

            for unprocessed_passenger in self._bus.processable_passengers:
                self._bus.ticket.log.append_text(
                    "Unprocessed passenger: " + unprocessed_passenger.passenger.id_text)
        except Exception as read_error: # pylint: disable=W0703
            self._log_exception(read_error)

    def _seat_passengers(self):
        try:
            self._bus.ticket.log.append_text("Writing new passengers to queue")

            if len(self._bus.new_passengers) <= 0:
                return

            self._bus.queue.insert(self._bus.new_passengers)
        except Exception as seat_error: # pylint: disable=W0703
            self._log_exception(seat_error)



""" Web module for exporting data """
import json
from typing import List
from flask import jsonify, render_template, request, redirect, url_for
from databus.client.client import Client
from databus.client.log import Log
from databus.database.export.export import ExporterFactory, ExportStatus
from databus.database.export.export_request import ExportRequest
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError


class ExportableClientList: # pylint: disable=R0903
    """ Class dealing with exportable clients """
    def __init__(self, p_dispatcher: AbstractDispatcher):
        self._dispatcher = p_dispatcher

    def get_exportable_clients(self, p_client_id: str = None) -> List[Client]:
        """ Returns a list of exportable clients """
        client_db = self._dispatcher.get_client_database(p_client_id)
        if p_client_id is None:
            return client_db.get_clients()
        return [client_db.client]


class ExportController(AbstractController):
    """ Export data """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_root(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        exporter = ExporterFactory.get_singleton()
        if exporter.status == ExportStatus.BUSY:
            return render_template("export_busy.html", alias=self.dispatcher.ticket.system_alias)

        list_builder = ExportableClientList(self.dispatcher)
        exportable_clients = list_builder.get_exportable_clients(self.authenticated_client_id)
        db_modules = self.dispatcher.ticket.database_factory.database_modules

        return render_template(
            "export.html",
            exportable_clients=exportable_clients,
            db_modules=db_modules,
            alias=self.dispatcher.ticket.system_alias)


class ExportGetDictController(AbstractController):
    """ Returns the dict of the database """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_root(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        db_module_name = request.args.get("db_module", 0, type=str)
        classes = self.dispatcher.ticket.database_factory.database_classes
        db_module = __import__(db_module_name, fromlist=classes)
        output = getattr(db_module, "ARGS_TEMPLATE")
        return jsonify(output)


class ExportExeController(AbstractController):
    """ Executes an export request """

    def execute(self):
        """ Executes an export request """
        try:
            self._authenticate(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        exporter = ExporterFactory.get_singleton()
        if exporter.status == ExportStatus.BUSY:
            return render_template("export_busy.html", alias=self.dispatcher.ticket.system_alias)

        client = request.form["client"]
        db_module = request.form["db_module"]
        db_args = json.loads(request.form["args"])
        mode = request.form["mode"]

        if self.authenticated_client_id != Client.ROOT:
            if client != self.authenticated_client_id:
                return redirect(url_for("_login"), code=302)

        target_db = self.dispatcher.ticket.database_factory.create_database(
            db_module,
            client,
            Log(),
            self.dispatcher.ticket.passenger_factory,
            db_args
        )

        export_request = ExportRequest(
            self.dispatcher,
            target_db,
            db_args,
            client,
            self.authenticated_client_id
        )

        if mode == "async":
            exporter.execute_async(export_request)
        else:
            exporter.execute(export_request)

        return render_template(
            "export_started.html",
            mode=mode,
            alias=self.dispatcher.ticket.system_alias)



""" Abstract passenger module """
from abc import ABC
from datetime import datetime
from typing import List
from uuid import uuid1, UUID
from zipfile import ZipFile
import os
import shutil
import mimetypes
from urlextract import URLExtract
import requests
from databus.passenger.attachment import Attachment, AttachmentFormat


class AbstractPassenger(ABC): # pylint: disable=R0903, R0902
    """ Abstract passenger class """
    _clock_seq: int = 0
    _TMP_ZIP_DIR = "_tmp_zip"
    _TMP_ZIP_EXTRACT_DIR = "_extracted"
    _TMP_ZIP_FILE = "_tmp_zip.zip"

    def __init__(self, # pylint: disable=R0913, R0912
                 p_external_id: str = None,
                 p_internal_id: UUID = None,
                 p_source_system: str = None,
                 p_attachments: List[Attachment] = None,
                 p_puller_module: str = None,
                 p_pull_datetime: datetime = None,
                 p_passenger_module: str = None,
                 p_log_guids: List[UUID] = None):

        if p_external_id is None:
            self.external_id = ""
        else:
            self.external_id = p_external_id

        if p_source_system is None:
            self.source_system = ""
        else:
            self.source_system = p_source_system

        if p_puller_module is None:
            self.puller_module = ""
        else:
            self.puller_module = p_puller_module

        if p_pull_datetime is None:
            self.pull_datetime = datetime.now()
        else:
            self.pull_datetime = p_pull_datetime

        if p_attachments is None:
            self.attachments = []
        else:
            self.attachments = p_attachments

        if p_internal_id is None:
            AbstractPassenger._clock_seq += 1
            self.internal_id = uuid1(clock_seq=AbstractPassenger._clock_seq)
        else:
            self.internal_id = p_internal_id

        if p_passenger_module is None:
            self.passenger_module = self.__module__
        else:
            self.passenger_module = p_passenger_module

        if p_log_guids is None:
            self._log_guids = []
        else:
            self._log_guids = p_log_guids

    @property
    def id_text(self) -> str:
        """ Returns the unique ID of the passenger as text """
        return self.source_system + " - " + self.external_id + " (" + str(self.internal_id) + ")"

    @property
    def log_guids(self) -> List[UUID]:
        """ Returns log guids """
        return self._log_guids

    def collect_log_guid(self, guid: UUID):
        """ Appends a new log guid, preventing duplicates """
        if guid in self._log_guids:
            return
        self._log_guids.append(guid)

    def get_attachment_by_name(self, name: str) -> Attachment:
        """ Returns the attachment """
        for attachment in self.attachments:
            if attachment.name == name:
                return attachment
        return None

    def unzip_attachments(self):
        """ Finds .ZIP attachments, and turns them into regular
        attachments within the same object
        """
        mime_type = mimetypes.MimeTypes()
        attachment_index = -1
        deletable_indices = []
        new_attachments = []

        for attachment in self.attachments:
            attachment_index += 1
            if len(attachment.name) < 4:
                continue
            if attachment.name.upper()[-4:] != ".ZIP":
                continue

            shutil.rmtree(AbstractPassenger._TMP_ZIP_DIR, ignore_errors=True)
            os.makedirs(AbstractPassenger._TMP_ZIP_DIR, exist_ok=True)

            extract_dir = os.path.join(
                AbstractPassenger._TMP_ZIP_DIR,
                AbstractPassenger._TMP_ZIP_EXTRACT_DIR)

            zip_path = os.path.join(AbstractPassenger._TMP_ZIP_DIR, AbstractPassenger._TMP_ZIP_FILE)
            with open(zip_path, "wb") as zip_file:
                zip_file.write(attachment.binary_content)

            extracted_files = []
            with ZipFile(zip_path) as zip_file:
                files_in_zip = zip_file.namelist()
                for file_in_zip in files_in_zip:
                    extract_path = os.path.join(extract_dir, file_in_zip)
                    zip_file.extract(file_in_zip, extract_dir)
                    extracted_files.append(extract_path)

            for extracted_path in extracted_files:
                mime = mime_type.guess_type(extracted_path)[0]
                file_format = Attachment.guess_format_by_mime_type(mime)
                file_in_zip = os.path.split(extracted_path)[1]

                if file_format == AttachmentFormat.text:
                    with open(extracted_path, "r") as extracted_file:
                        file_content = extracted_file.read()
                        file_content = file_content.replace("\n", " ")
                        file_content = file_content.replace("\t", " ")
                    unzip_attachment = Attachment(
                        p_name=file_in_zip,
                        p_format=AttachmentFormat.text,
                        p_text_content=file_content)
                else:
                    with open(extracted_path, "rb") as extracted_file:
                        file_content = extracted_file.read()
                    unzip_attachment = Attachment(
                        p_name=file_in_zip,
                        p_format=AttachmentFormat.binary,
                        p_binary_content=file_content)

                new_attachments.append(unzip_attachment)

            deletable_indices.append(attachment_index)
            shutil.rmtree(AbstractPassenger._TMP_ZIP_DIR, ignore_errors=True)

        deletable_indices.sort(reverse=True)
        for deletable_index in deletable_indices:
            self.attachments.pop(deletable_index)

        for new_attachment in new_attachments:
            self.attachments.append(new_attachment)

    def download_links_in_html_as_attachments(self, p_html: str, p_extensions: List[str]):
        """ Scans the given HTML file, finds links, downloads
        the files and saves them as attachments.
        This method supports only text attachments at this time.
        """
        # Build clean HTML
        if p_html is None or len(p_html) <= 0:
            return
        clean_html = p_html.replace("\r", "").replace("\n", "")
        html_tag_pos = clean_html.lower().find("<html")
        if html_tag_pos < 0:
            return
        clean_html = clean_html[html_tag_pos:]

       # Extract URL's
        extractor = URLExtract()
        urls = extractor.find_urls(clean_html)

        # Download as necessary
        for url in urls:
            low_url = url.lower()
            has_eligible_extension = False
            for extension in p_extensions:
                low_extension = "." + extension.lower()
                if low_extension in low_url:
                    has_eligible_extension = True
                    break
            if not has_eligible_extension:
                continue

            if "urldefense.com" in url:
                real_http_pos = low_url.rfind("http")
                clean_url = url[real_http_pos:].replace("__", "")
            else:
                clean_url = url
            if clean_url[-1] == "/":
                clean_url = clean_url[:-1]

            filename = os.path.basename(clean_url)
            dummy_name, extension = os.path.splitext(filename)
            extension = extension.replace(".", "")
            file_format = Attachment.guess_format_by_file_extension(extension)

            response = requests.get(clean_url, allow_redirects=True)

            if file_format == AttachmentFormat.text:
                downloaded_attachment = Attachment(
                    p_name=filename,
                    p_format=AttachmentFormat.text,
                    p_text_content=response.text)
            else:
                downloaded_attachment = Attachment(
                    p_name=filename,
                    p_format=AttachmentFormat.binary,
                    p_binary_content=response.content)

            self.attachments.append(downloaded_attachment)



""" Module for deciding on SQL actions """
from typing import List
from databus.client.client import Client
from databus.database.difference_check import DifferenceChecker, TableKey, Action
from databus.database.sql_db.delete_builder import DeleteBuilder
from databus.database.sql_db.insert_builder import InsertBuilder
from databus.database.sql_db.update_builder import UpdateBuilder
from databus.database.sql_db.query_helper import QueryHelper
from databus.database.sql_db.table import Table


class ActionDecision:
    """ Action decision class
    Basically; this is the output of ActionDecider
    """
    def __init__(self, p_query_helper: QueryHelper):
        self.inserts = []
        self.updates = []
        self.deletes = []
        self._query_helper = p_query_helper

    def add_delete(self, p_table: str, p_row: dict):
        """ Adds a new delete command """
        delete = DeleteBuilder(
            self._query_helper.args,
            self._get_modifiable_client_id(p_table, p_row))
        delete.table = p_table
        key_fields = Table(self._query_helper, p_table).key_fields

        for key_field in key_fields:
            cell_value = p_row[key_field]
            if key_field == "client_id":
                self._validate_client_id(cell_value)
                continue
            if isinstance(cell_value, int):
                delete.where.add_and_field_eq_int(key_field, cell_value)
            else:
                delete.where.add_and_field_eq_str(key_field, cell_value)

        self.deletes.append(delete)

    def add_insert(self, p_table: str, p_row: dict):
        """ Adds a new insert command """
        insert = InsertBuilder(
            self._query_helper.args,
            self._get_modifiable_client_id(p_table, p_row))
        insert.table = p_table

        for table_col in p_row:
            cell_value = p_row[table_col]
            if table_col == "client_id":
                self._validate_client_id(cell_value)
                continue
            if isinstance(cell_value, int):
                insert.add_int(table_col, cell_value)
            else:
                insert.add_string(table_col, cell_value)

        self.inserts.append(insert)

    def add_update(self, p_table: str, p_row: dict):
        """ Adds a new update command """
        update = UpdateBuilder(
            self._query_helper.args,
            self._get_modifiable_client_id(p_table, p_row))

        update.table = p_table
        key_fields = Table(self._query_helper, p_table).key_fields

        for table_col in p_row:
            cell_value = p_row[table_col]
            if table_col == "client_id":
                self._validate_client_id(cell_value)
                continue

            if table_col in key_fields:
                if isinstance(cell_value, int):
                    update.where.add_and_field_eq_int(table_col, cell_value)
                else:
                    update.where.add_and_field_eq_str(table_col, cell_value)
            else:
                if isinstance(cell_value, int):
                    update.add_int(table_col, cell_value)
                else:
                    update.add_string(table_col, cell_value)

        self.updates.append(update)

    def _get_modifiable_client_id(self, p_table: str, p_row: dict):
        if p_table == "client" and self._query_helper.client_id == Client.ROOT:
            return p_row["client_id"]
        return self._query_helper.client_id

    def _validate_client_id(self, p_client_id: str):
        if self._query_helper.client_id == Client.ROOT:
            return
        assert p_client_id == self._query_helper.client_id


class ActionDecider: # pylint: disable=R0903
    """ Action decider class
    This class will take 2 dictionaries.
    A) Dictionary in memory
    B) Dictionary in database
    It will compare both, and tell what needs to be inserted / updated / deleted
    """
    def __init__(self, p_query_helper: QueryHelper):
        self._memory = {}
        self._database = {}
        self._query_helper = p_query_helper
        self._decision = ActionDecision(self._query_helper)

    def decide(self, p_memory: dict, p_database: dict) -> ActionDecision:
        """ Decides what action(s) to take """
        self._memory = p_memory
        self._database = p_database
        self._decision = ActionDecision(self._query_helper)

        table_keys = self._get_table_keys(p_database)
        diff_checker = DifferenceChecker(table_keys, p_memory, p_database)

        for difference in diff_checker.result:
            if difference.action == Action.INSERT:
                self._decision.add_insert(difference.table, difference.row)
            elif difference.action == Action.UPDATE:
                self._decision.add_update(difference.table, difference.row)
            elif difference.action == Action.DELETE:
                self._decision.add_delete(difference.table, difference.row)

        return self._decision

    def decide_and_execute(self, p_memory: dict, p_database: dict):
        """ Decides what action(s) to take, and executes them """
        decision = self.decide(p_memory, p_database)

        try:
            action_taken = False

            for insert in decision.inserts:
                self._query_helper.execute_insert(insert)
                action_taken = True

            for update in decision.updates:
                self._query_helper.execute_update(update)
                action_taken = True

            for delete in decision.deletes:
                self._query_helper.execute_delete(delete)
                action_taken = True

            if action_taken:
                self._query_helper.commit()

        except Exception as error:
            self._query_helper.rollback()
            raise error

    def _get_table_keys(self, p_database: dict) -> List[TableKey]:
        output = []
        for table_name in p_database:
            table_obj = Table(self._query_helper, table_name)
            table_keys = table_obj.key_fields
            table_key_obj = TableKey(table_name, table_keys)
            output.append(table_key_obj)
        return output



""" SQL Server database argument module
This module has imitated the Sql database argument module
"""
from enum import Enum

class SqlDatabaseArgumentError(Exception):
    """ Sql database argument exception """

    class ErrorCode(Enum):
        """ Error code """
        invalid_argument: 1

    def __init__(self, p_error_code: ErrorCode, p_argument: str = ""):
        super().__init__()
        self.error_code = p_error_code

        if p_argument is None:
            self.argument = ""
        else:
            self.argument = p_argument

    @property
    def message(self) -> str:
        """ Error message as string """
        if self.error_code == SqlDatabaseArgumentError.ErrorCode.invalid_argument:
            return "Invalid SqlDatabase argument: " + self.argument
        return "SqlDatabase argument error"


class SqlDatabaseArguments: # pylint: disable=R0902, R0903
    """ Sql database argument class """
    KEY_SERVER = "server"
    KEY_DATABASE = "database"
    KEY_USERNAME = "username"
    KEY_PASSWORD = "password"
    KEY_SCHEMA = "schema"

    TEMPLATE = {
        KEY_SERVER: "(server address)",
        KEY_DATABASE: "(database name)",
        KEY_USERNAME: "(username)",
        KEY_PASSWORD: "(password)",
        KEY_SCHEMA: "(database schema)"
    }

    def __init__(self, p_arguments: dict):
        for key in p_arguments:
            if key == SqlDatabaseArguments.KEY_SERVER:
                self.server = p_arguments[key]
            elif key == SqlDatabaseArguments.KEY_DATABASE:
                self.database = p_arguments[key]
            elif key == SqlDatabaseArguments.KEY_USERNAME:
                self.username = p_arguments[key]
            elif key == SqlDatabaseArguments.KEY_PASSWORD:
                self.password = p_arguments[key]
            elif key == SqlDatabaseArguments.KEY_SCHEMA:
                self.schema = p_arguments[key]
            else:
                raise SqlDatabaseArgumentError(
                    SqlDatabaseArgumentError.ErrorCode.invalid_argument,
                    p_argument=key)



""" Module to pull passenger from multiple Exchange accounts

If the expected E-Mails land to a single Exchange account, you can use
databus.puller.abstract_exchange.

If the expected E-Mail can land on multiple Exchange accounts, you can use
this module.
"""
from abc import ABC, abstractproperty
import json
from typing import List
from databus.client.log import Log
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.passenger.attachment import Attachment
from databus.puller.abstract_puller import AbstractPuller
from databus.puller.abstract_exchange import AbstractExchange

class ExchangeAccount:
    """ Defines a singular Exchange account """
    def __init__(self, p_puller: AbstractExchange = None, p_alias: str = None):
        self.puller = p_puller
        self.alias = p_alias

class AbstractMultiExchange(AbstractPuller, ABC):
    """ Class to pull passenger from multiple Exchange accounts """
    _ARTIFICIAL_ATTACHMENT_FILE = "_databus_abstract_multi_exchange.json"

    def __init__(self, p_log: Log = None):
        super().__init__(p_log=p_log)

    @abstractproperty
    def exchange_accounts(self) -> List[ExchangeAccount]:
        """ Returns a list of Exchange accounts """

    def notify_passengers_seated(self, p_seated_passengers: List[AbstractPassenger]):
        """ Called after a passenger is properly queued.
        You would typically write a code here to ensure that the passenger is not
        returned any more when the puller works again.
        """
        for exchange_account in self.exchange_accounts:
            seated_passengers_of_account = []
            for seated_passenger in p_seated_passengers:
                artificial_attachment = seated_passenger.get_attachment_by_name(
                    AbstractMultiExchange._ARTIFICIAL_ATTACHMENT_FILE)
                aa_dict = json.loads(artificial_attachment.text_content)
                if aa_dict["alias"] == exchange_account.alias:
                    seated_passengers_of_account.append(seated_passenger)
            if len(seated_passengers_of_account) > 0:
                exchange_account.puller.notify_passengers_seated(seated_passengers_of_account)

    def pull(self) -> List[AbstractPassenger]:
        """ Pulls passengers from the source system """
        output = []
        for exchange_account in self.exchange_accounts:
            new_emails = exchange_account.puller.pull()
            for new_email in new_emails:
                artificial_attachment_dict = {"alias": exchange_account.alias}
                artificial_attachment_json = json.dumps(artificial_attachment_dict)
                artificial_attachment = Attachment(
                    p_name=AbstractMultiExchange._ARTIFICIAL_ATTACHMENT_FILE,
                    p_text_content=artificial_attachment_json)
                new_email.attachments.append(artificial_attachment)
                output.append(new_email)
        return output



""" Observer module for significant dispatcher events """
from abc import ABC, abstractmethod
from databus.client.client import Client
from databus.client.client_passenger import ClientPassenger
from databus.client.log import Log


class DispatcherObserver(ABC): # pylint: disable=R0903
    """ Observer class for significant dispatcher events
    You can implement this class into your project and pass it via
    the dispatcher ticket.
    """
    @abstractmethod
    def drive_passenger_complete(self,
                                 p_client: Client,
                                 p_client_passenger: ClientPassenger,
                                 p_log: Log):
        """ Fired when primal_dispatcher.drive_passenger is completed.
        You can typically use this method to send E-Mail notifications if p_log contains errors.
        """



""" Demo puller module returning binary attachments """
from os import path
from typing import List
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.passenger.demo.demo_binary_passenger import DemoBinaryPassenger
from databus.passenger.attachment import Attachment, AttachmentFormat
from databus.puller.abstract_puller import AbstractPuller

class DemoBinaryPuller(AbstractPuller):
    """ Demo binary puller class """
    _BIN_FILE_NAME = "sample_binary_file.png"
    _MODULE_FILE_NAME = "demo_binary_puller.py"

    def notify_passengers_seated(self, p_seated_passengers: List[AbstractPassenger]):
        """ Handles passengers which are queued already """
        for seated_passenger in p_seated_passengers:
            self.log.append_text(
                "Demo binary puller notified about seated passenger " + seated_passenger.id_text)

    def pull(self) -> List[DemoBinaryPassenger]:
        """ Fake pull from imaginary source system """
        output = []
        passenger1 = DemoBinaryPassenger()
        passenger1.external_id = "ID_BIN_1"
        passenger1.source_system = "DEMO_SYSTEM"
        passenger1.puller_module = self.__module__
        passenger1.attachments.append(Attachment(
            p_name=DemoBinaryPuller._BIN_FILE_NAME,
            p_format=AttachmentFormat.binary,
            p_binary_content=DemoBinaryPuller._get_sample_binary()))
        output.append(passenger1)
        self.log.append_text("Got passenger " + passenger1.id_text)
        return output

    @staticmethod
    def _get_sample_binary() -> bytearray:
        """ Returns the binary content of the sample file """
        full_path = path.abspath(__file__).replace(
            DemoBinaryPuller._MODULE_FILE_NAME,
            DemoBinaryPuller._BIN_FILE_NAME)
        with open(full_path, "rb") as bin_file:
            binary_content = bin_file.read()
        return binary_content



""" Module for queue controllers """
from datetime import datetime
from typing import List
from flask import redirect, render_template, request, url_for
from databus.client.client import Client
from databus.client.client_passenger import ClientPassenger
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher
from databus.pqueue.queue_status import PassengerQueueStatus, QueueStatus
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError


class ClientPassengerQueue: # pylint: disable=R0903
    """ Defines a client passenger and all queue files present """
    def __init__(self,
                 p_client: Client = None,
                 p_client_passenger: ClientPassenger = None,
                 p_queues: List[PassengerQueueStatus] = None):

        if p_client is None:
            self.client = Client()
        else:
            self.client = p_client

        if p_client_passenger is None:
            self.client_passenger = []
        else:
            self.client_passenger = p_client_passenger

        if p_queues is None:
            self.queues = []
        else:
            self.queues = p_queues


class ClientPassengerQueueReader:
    """ Class dealing with client queue files """
    def __init__(self, p_dispatcher: AbstractDispatcher):
        self._dispatcher = p_dispatcher

    def get_client_passenger_queue_entry(self,
                                         p_client_id: str,
                                         p_internal_id: str
                                        ) -> PassengerQueueStatus:
        """ Returns a list of client - passenger - queue files """
        client_database = self._dispatcher.get_client_database(p_client_id)
        return client_database.get_passenger_queue_entry(p_internal_id)

    def get_client_passenger_queue_list(self,
                                        p_client_id: str = None
                                       ) -> List[ClientPassengerQueue]:
        """ Returns a list of client - passenger - queue files """
        output = []
        for client in self._dispatcher.all_clients:
            if p_client_id is not None and client.id != p_client_id:
                continue
            client_database = self._dispatcher.get_client_database(client.id)
            for client_passenger in client.passengers:
                queue_entries = client_database.get_passenger_queue_entries(p_passenger_module=client_passenger.name) # pylint: disable=C0301

                client_passenger_queue = ClientPassengerQueue(
                    p_client=client,
                    p_client_passenger=client_passenger,
                    p_queues=queue_entries)

                output.append(client_passenger_queue)
        return output


class QueueAttachmentController(AbstractController):
    """ Queue attachment download """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        passenger = request.args.get("passenger", 0, type=str)
        file = request.args.get("file", 0, type=str)
        entry = ClientPassengerQueueReader(self.dispatcher).get_client_passenger_queue_entry(self.requested_client_id, passenger) # pylint: disable=C0301

        for att in entry.passenger.attachments:
            if att.name == file:
                return AbstractController._download_attachment(att)

        return "File not found"


class QueueDisplayController(AbstractController):
    """ Queue display """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        passenger = request.args.get("passenger", 0, type=str)
        entry = ClientPassengerQueueReader(self.dispatcher).get_client_passenger_queue_entry(self.requested_client_id, passenger) # pylint: disable=C0301

        log_ids = []
        client_database = self.requested_client_database
        for log_guid in entry.passenger.log_guids:
            log_id = client_database.convert_log_guid_to_id(log_guid)
            log_ids.append(log_id)

        return render_template(
            "queue_display.html",
            client=self.requested_client_id,
            entry=entry,
            alias=self.dispatcher.ticket.system_alias,
            log_ids=log_ids)


class QueueListController(AbstractController):
    """ Queue list """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_root()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        queue_reader = ClientPassengerQueueReader(self.dispatcher)
        entries = queue_reader.get_client_passenger_queue_list(
            p_client_id=self.authenticated_client_id)

        return render_template(
            "queue_list.html",
            entries=entries,
            alias=self.dispatcher.ticket.system_alias)


class QueuePurgeController(AbstractController):
    """ Queue purge """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        passenger_module = request.args.get("passenger", 0, type=str)
        client_db = self.requested_client_database
        client_passenger = client_db.client.get_client_passenger(passenger_module)

        queue = self.dispatcher.ticket.queue_factory.create_queue(
            client_passenger.queue_module,
            client_db,
            client_db.log)

        queue.delete_completed_passengers(passenger_module, datetime.now())
        return redirect(url_for("_queue_list"), code=302)


class QueueStatusUpdateController(AbstractController):
    """ Queue display """

    def execute(self): # pylint: disable=R0912
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        passenger = request.args.get("passenger", "", type=str)
        entry = ClientPassengerQueueReader(self.dispatcher).get_client_passenger_queue_entry(self.requested_client_id, passenger) # pylint: disable=C0301

        puller_notified = request.args.get("puller_notified", "", type=str)
        if puller_notified == "true":
            entry.puller_notified = True
        elif puller_notified == "false":
            entry.puller_notified = False

        processor_module = request.args.get("processor", "", type=str)
        if processor_module != "":
            processed = request.args.get("processed", "", type=str) == "true"
            for processor_status in entry.processor_statuses:
                if processor_status.processor_module == processor_module:
                    if processed:
                        processor_status.status = QueueStatus.complete
                    else:
                        processor_status.status = QueueStatus.incomplete

        pusher_module = request.args.get("pusher", "", type=str)
        if pusher_module != "":
            processed = request.args.get("pushed", "", type=str) == "true"
            for pusher_status in entry.pusher_statuses:
                if pusher_status.pusher_module == pusher_module:
                    if processed:
                        pusher_status.status = QueueStatus.complete
                    else:
                        pusher_status.status = QueueStatus.incomplete

        self.requested_client_database.update_queue_status(entry)

        redirect_url = url_for("_queue_display")
        redirect_url += "?client=" + self.requested_client_id + "&passenger=" + passenger
        redirect_url += AbstractController._get_cache_buster()
        return redirect(redirect_url, code=302)



""" Implementation module for JSON database """
from datetime import datetime
from typing import List
from uuid import UUID
from databus.client.client import Client
from databus.client.log import Log, MessageType
from databus.client.user import Credential
from databus.database.abstract_database import AbstractDatabase, LogListItem
from databus.database.json_db.json_client import JsonClient
from databus.database.json_db.json_database_arguments import JsonDatabaseArguments
from databus.database.json_db.json_log import JsonLog
from databus.database.json_db.json_queue import JsonQueue
from databus.passenger.abstract_factory import AbstractPassengerFactory
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.pqueue.queue_status import QueueStatus, PassengerQueueStatus


ARGS_TEMPLATE = JsonDatabaseArguments.TEMPLATE


class JsonDatabase(AbstractDatabase):
    """ Implementation class for JSON database """
    def __init__(self,
                 p_client_id: str,
                 p_log: Log,
                 p_passenger_factory: AbstractPassengerFactory,
                 p_arguments: dict):
        super().__init__(p_client_id, p_log, p_passenger_factory, p_arguments)
        self._args = JsonDatabaseArguments(p_arguments)
        self._json_client = JsonClient(self._args)
        if p_client_id is None or p_client_id == "":
            self.client = None
        else:
            try:
                self.client = self._get_client(p_client_id)
            except Exception: # pylint: disable=W0703
                self.client = None # We might be creating a new database
        self._json_log = JsonLog(self._args)
        self._json_queue = JsonQueue(p_client_id, p_log, self.passenger_factory, self._args)

    @property
    def client_master_data(self) -> str:
        """ Returns the definition of all clients as a JSON string """
        return self._json_client.client_master_as_json

    @client_master_data.setter
    def client_master_data(self, p_definitions: str):
        """ Sets the definitions into the clients config JSON """
        self._json_client.client_master_as_json = p_definitions

    @property
    def customizing(self) -> str:
        """ Returns the client customizing as JSON """
        return self._json_client.get_config_as_string(self.client_id)

    @customizing.setter
    def customizing(self, p_customizing: str):
        """ Sets the customizing into the clients config JSON """
        self._json_client.save_config(self.client_id, p_customizing)

    def delete_old_logs(self, p_before: datetime):
        """ Deletes old logs from the disk """
        self.log.append_text("Deleting logs before " + p_before.isoformat())
        self._json_log.delete_log_file_before(self.client.id, p_before, self.log)

    def delete_passenger_queue(self, p_passengers: List[AbstractPassenger]):
        """ Deletes passenger from disk """
        self.log.append_text("Deleting passengers from queue")
        self._json_queue.delete_passengers(p_passengers)

    def ensure_schema_existence(self):
        """ Checks the schema for the client
        Creates / completes the schema if anything is missing
        """
        self._json_client.ensure_schema_existence(self.client_id)

    def erase_passenger_queue(self):
        """ Deletes all passengers from disk """
        self.log.append_text("Erasing passenger queue")
        self._json_queue.erase_passenger_queue()

    def get_clients(self) -> List[Client]:
        """ Returns a list of clients """
        return self._json_client.all_clients

    def get_log_content(self, p_log_id: str) -> str:
        """ Returns the contents of the given log
        p_log_id is whatever you have returned in get_log_list.
        """
        return self._json_log.get_log_file_content(self.client_id, p_log_id)

    def get_log_list(self) -> List[LogListItem]:
        """ Returns log entries """
        output = []
        b_warning = "[" + MessageType.warning.name + "]"
        b_error = "[" + MessageType.error.name + "]"

        log_ids = self._json_log.get_log_file_list(self.client_id)
        for log_id in log_ids:
            output_item = LogListItem(p_log_id=log_id)
            log_content = self.get_log_content(log_id)
            if b_error in log_content:
                output_item.worst_message_type = MessageType.error
            elif b_warning in log_content:
                output_item.worst_message_type = MessageType.warning
            else:
                output_item.worst_message_type = MessageType.info
            output.append(output_item)
        return output

    def get_passenger_queue_entries(self, # pylint: disable=R0913
                                    p_passenger_module: str = None,
                                    p_processor_status: QueueStatus = None,
                                    p_pusher_status: QueueStatus = None,
                                    p_puller_notified: bool = None,
                                    p_pulled_before: datetime = None
                                    ) -> List[PassengerQueueStatus]:
        """ Reads the requested entries from the disk """
        self.log.append_text("Reading passenger queue entries")
        return self._json_queue.get_passengers(p_passenger_module,
                                               p_processor_status,
                                               p_pusher_status,
                                               p_puller_notified,
                                               p_pulled_before)

    def get_passenger_queue_entry(self, # pylint: disable=R0913
                                  p_internal_id: str
                                 ) -> PassengerQueueStatus:
        return self._json_queue.get_passengers(p_internal_id=p_internal_id)[0]

    def insert_client(self, p_client: Client):
        """ Inserts a new client """
        self._json_client.create_client_if_missing__by_obj(p_client)

    def insert_log(self, p_log: Log):
        """ Creates a new log file on the disk """
        self.log.append_text("Writing log to disk")
        if self.client is None:
            return
        self._json_log.insert(self.client.id, p_log)

    def insert_passenger_queue(self, p_passenger_status: PassengerQueueStatus):
        """ Writes new files to the disk """
        self.log.append_text("Appending passenger " + p_passenger_status.passenger.id_text)
        self._json_queue.insert_passenger(p_passenger_status)

    def update_queue_status(self, p_status: PassengerQueueStatus):
        """ Updates queue files on the disk """
        self.log.append_text("Updating passenger " + p_status.passenger.id_text)
        self._json_queue.update_passenger(p_status)

    def update_user_credential(self, p_credential: Credential):
        """ Updates the credential of the given user """
        self._json_client.update_user_credential(self.client_id, p_credential)

    def convert_log_guid_to_id(self, p_guid: UUID) -> str:
        """ UUID to id conversion """
        guid_as_str = str(p_guid)
        log_ids = self._json_log.get_log_file_list(self.client_id)
        for log_id in log_ids:
            if guid_as_str in log_id:
                return log_id
        return ""

    def _get_client(self, p_id: str) -> Client:
        return self._json_client.get_single(p_id)



""" Module for abstract controller
A controller actually corresponds to a method in the
app.py file of Flask. It is named after MVC controllers
because the purpose is the same.
"""
from abc import ABC, abstractmethod
import io
import uuid
from flask import redirect, request, send_file, session, url_for
from databus.client.client import Client
from databus.client.user import User
from databus.database.abstract_database import AbstractDatabase
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher
from databus.passenger.attachment import Attachment, AttachmentFormat
from databus.web.util import user_is_admin


class ClientUser: # pylint: disable=R0903
    """ Defines a user of a client """
    def __init__(self, p_client: Client = None, p_user: User = None):
        if p_client is None:
            self.client = Client()
        else:
            self.client = p_client
        if p_user is None:
            self.user = User()
        else:
            self.user = p_user


class ClientUserFinder: # pylint: disable=R0903
    """ Class to find client users """
    def __init__(self, p_dispatcher: AbstractDispatcher):
        self._dispatcher = p_dispatcher

    def find_by_token(self, p_token: str) -> ClientUser:
        """ Locates & returns a user by his/her token """
        all_clients = self._dispatcher.all_clients
        for client in all_clients:
            for user in client.users:
                if user.credential.token == p_token:
                    return ClientUser(p_client=client, p_user=user)
        return None


class AuthenticationError(Exception):
    """ Dispatcher creation exception """
    def __init__(self, output):
        super().__init__()
        self.output = output


class AbstractController(ABC):
    """ Base class for any controller """
    def __init__(self, p_dispatcher: AbstractDispatcher):
        self.dispatcher = p_dispatcher
        self.authenticated_client_id = None
        self.requested_client_id = ""

    @property
    def authenticated_client_database(self) -> AbstractDatabase:
        """ Returns the database object of the requested client """
        return self.dispatcher.get_client_database(self.authenticated_client_id)

    @property
    def requested_client_database(self) -> AbstractDatabase:
        """ Returns the database object of the requested client """
        return self.dispatcher.get_client_database(self.requested_client_id)

    @abstractmethod
    def execute(self):
        """ Builds and returns the page """

    @staticmethod
    def _download_attachment(att: Attachment):
        if att is None:
            return "File not found"
        if att.format == AttachmentFormat.binary:
            return send_file(
                io.BytesIO(att.binary_content),
                attachment_filename=att.name,
                as_attachment=True)
        if att.format == AttachmentFormat.text:
            return att.text_content
        return "Unexpected attachment format"

    @staticmethod
    def _get_cache_buster() -> str:
        return "&cache_buster=" + str(uuid.uuid1())

    def _authenticate(self, must_be_admin=False):
        self.authenticated_client_id = self._get_authenticated_client_id()
        if self.authenticated_client_id is None:
            output = redirect(url_for("_login"), code=302)
            raise AuthenticationError(output)
        if must_be_admin and (not user_is_admin()):
            output = redirect(url_for("_home"), code=302)
            raise AuthenticationError(output)

    def _authenticate_minding_requested_client(self, must_be_admin=False):
        self._authenticate(must_be_admin)
        self.requested_client_id = request.args.get("client", "", type=str)
        if self.requested_client_id == "":
            try:
                self.requested_client_id = request.form["client"]
            except Exception: # pylint: disable=W0703
                self.requested_client_id = ""

        if self.authenticated_client_id not in (Client.ROOT, self.requested_client_id):
            output = redirect(url_for("_home"), code=302)
            raise AuthenticationError(output)

    def _authenticate_minding_root(self, must_be_admin=False):
        self._authenticate(must_be_admin)
        if self.authenticated_client_id == Client.ROOT:
            self.authenticated_client_id = None

    def _get_authenticated_client_id(self) -> str:
        if "client_id" in session and session["client_id"] != "":
            if self._is_client_id_valid(session["client_id"]):
                return session["client_id"]

        token = request.cookies.get("token")
        if token != "":
            client_user = ClientUserFinder(self.dispatcher).find_by_token(token)
            if client_user is not None:
                session["client_id"] = client_user.client.id
                session["user_role"] = client_user.user.role.name
                return client_user.client.id

        return None

    def _is_client_id_valid(self, p_client_id: str):
        for client in self.dispatcher.all_clients:
            if client.id == p_client_id:
                return True
        return False



""" Attachment module
All settings, data, files, etc of a passengers are
considered to be separate attachment files.
If you need to store some settings and definitions
per passenger, consider creating a JSON attachment.
"""
from enum import Enum
from typing import List
from os import path


class AttachmentError(Exception):
    """ Attachment exception class """

    class ErrorCode(Enum):
        """ Attachment error code """
        invalid_format: 1
        invalid_name: 2
        duplicate_name: 3

    def __init__(self, p_error_code: ErrorCode, p_format: str = None, p_name: str = None):
        super().__init__()
        self.error_code = p_error_code

        if p_format is None:
            self.format = ""
        else:
            self.format = p_format

        if p_name is None:
            self.name = ""
        else:
            self.name = p_name

    @property
    def message(self) -> str:
        """ Attachment error text """
        if self.ErrorCode == AttachmentError.ErrorCode.invalid_format:
            return "Invalid attachment format: " + self.format
        if self.ErrorCode == AttachmentError.ErrorCode.invalid_name:
            return "Invalid attachment name: " + self.name
        if self.ErrorCode == AttachmentError.ErrorCode.duplicate_name:
            return "Duplicate attachment name: " + self.name
        return "Attachment error"


class AttachmentFormat(Enum):
    """ Attachment format type """
    text = 1
    binary = 2


class Attachment: # pylint: disable=R0903
    """ Attachment class """

    _VALID_CHARS = "1234567890QWERTYUIOPASDFGHJKLZXCVBNMqwertyuiopasdfghjklzxcvbnm_."
    _REPLACEMENT_CHAR = "_"

    def __init__(self,
                 p_name: str = None,
                 p_format: AttachmentFormat = AttachmentFormat.text,
                 p_text_content: str = None,
                 p_binary_content: bytearray = None):

        Validator.validate_attachment_format(p_format)

        if p_name is None:
            self._name = ""
        else:
            self._name = Attachment._cleanse_name(p_name)

        self.format = p_format

        if p_text_content is None:
            self.text_content = ""
        else:
            self.text_content = p_text_content

        self.binary_content = p_binary_content

    @property
    def name(self) -> str:
        """ Attachment name """
        return self._name

    @name.setter
    def name(self, p_val: str):
        """ Attachment name setter
        We cleanse the attachment name from invalid characters here
        """
        self._name = Attachment._cleanse_name(p_val)

    @staticmethod
    def guess_format_by_mime_type(p_mime_type: str) -> AttachmentFormat:
        """ Returns if attachment is text or binary """
        if any(["text" in p_mime_type,
                "txt" in p_mime_type,
                "html" in p_mime_type,
                "json" in p_mime_type,
                "application/xml" in p_mime_type]):
            return AttachmentFormat.text
        return AttachmentFormat.binary

    @staticmethod
    def guess_format_by_file_extension(p_extension: str) -> AttachmentFormat:
        """ Returns if attachment is text or binary """
        low_ext = p_extension.lower().replace(".", "")
        if low_ext in ["txt", "html", "json", "xml"]:
            return AttachmentFormat.text
        return AttachmentFormat.binary

    @staticmethod
    def guess_format_by_file_name(p_name: str) -> AttachmentFormat:
        """ Returns if attachment is text or binary """
        name_split = path.splitext(p_name)
        if len(name_split) < 2:
            return Attachment.guess_format_by_file_extension("")
        return Attachment.guess_format_by_file_extension(name_split[1])

    @staticmethod
    def _cleanse_name(p_val: str) -> str:
        output = ""
        for name_char in p_val:
            if name_char in Attachment._VALID_CHARS:
                output += name_char
            else:
                output += Attachment._REPLACEMENT_CHAR
        return output


class Validator:
    """ Attachment validator class """
    @staticmethod
    def ensure_all_names_are_unique(p_attachments: List[Attachment], p_correct: bool = True):
        """ Prevents duplicate file names among attachments """
        if p_correct:
            iter_count = 0
            while True:
                try:
                    Validator._enforce_names_are_unique(p_attachments)
                    return
                except Exception as error: # pylint: disable=W0703
                    iter_count += 1
                    if iter_count > 30:
                        raise error
                    Validator._rename_duplicates(p_attachments)
        else:
            Validator._enforce_names_are_unique(p_attachments)

    @staticmethod
    def _enforce_names_are_unique(p_attachments: List[Attachment]):
        name_count = {}
        for attachment in p_attachments:
            if attachment.name in name_count:
                name_count[attachment.name] += 1
            else:
                name_count[attachment.name] = 1

        for name in name_count:
            if name_count[name] > 1:
                raise AttachmentError(AttachmentError.ErrorCode.duplicate_name, p_name=name)

    @staticmethod
    def _rename_duplicates(p_attachments: List[Attachment]):
        unique_names = []
        for attachment in p_attachments:
            if attachment.name not in unique_names:
                unique_names.append(attachment.name)

        for name in unique_names:
            name_count = 0
            for attachment in p_attachments:
                if attachment.name != name:
                    continue
                name_count += 1
                if name_count < 2:
                    continue
                att_name, att_extension = path.splitext(attachment.name)
                att_name += " (" + str(name_count) + ")"
                attachment.name = att_name + att_extension

    @staticmethod
    def validate_attachment_format(p_format: AttachmentFormat):
        """ Validates attachment format code """
        if p_format is None or p_format not in AttachmentFormat:
            raise AttachmentError(AttachmentError.ErrorCode.invalid_format, p_format=p_format)

    @staticmethod
    def validate_attachment_name(p_name: str):
        """ Validates attachment file name """
        if p_name is None or p_name == "":
            raise AttachmentError(AttachmentError.ErrorCode.invalid_name, p_name=p_name)

    @staticmethod
    def validate_attachments(p_attachments: List[Attachment]):
        """ Runs all validations for attachments """
        for attachment in p_attachments:
            Validator.validate_attachment_name(attachment.name)
            Validator.validate_attachment_format(attachment.format)
        Validator.ensure_all_names_are_unique(p_attachments)



""" Module for data export request """
from databus.database.abstract_database import AbstractDatabase
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher

class ExportRequest: # pylint: disable=R0903, R0913
    """ Defines an export request """

    def __init__(self,
                 p_dispatcher: AbstractDispatcher,
                 p_to: AbstractDatabase,
                 p_args: dict,
                 p_client: str,
                 p_requesting_client_id: str):
        self.target_db = p_to
        self.args = p_args
        self.client = p_client
        self.dispatcher = p_dispatcher
        self.requesting_client_id = p_requesting_client_id



""" Default queue module """
from datetime import datetime
from typing import List
from databus.client.log import LogEntry, MessageType
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.pqueue.abstract_queue import AbstractQueue
from databus.pqueue.queue_status import QueueStatus, PassengerQueueStatus, QueueStatusFactory


class PrimalQueue(AbstractQueue):
    """ Default queue class """

    def delete_completed_passengers(self, p_passenger_module: str, p_pulled_before: datetime):
        """ Deletes completed passengers from the database,
        which were pulled before p_pulled_before and queued, processed, puller-notified, pushed
        """
        self.log.append_text("Deleting " +
                             p_passenger_module +
                             " passengers completed before " +
                             p_pulled_before.isoformat())

        deletable_queue_entries = self.database.get_passenger_queue_entries(
            p_passenger_module=p_passenger_module,
            p_processor_status=QueueStatus.complete,
            p_pusher_status=QueueStatus.complete,
            p_puller_notified=True,
            p_pulled_before=p_pulled_before)

        deletable_passengers = []

        for queue_entry in deletable_queue_entries:
            self.log.append_text(
                "Initiating deletion for completed passenger " + queue_entry.passenger.id_text)
            deletable_passengers.append(queue_entry.passenger)

        self.database.delete_passenger_queue(deletable_passengers)

    def erase(self):
        """ Deletes all passengers from the database """
        self.database.erase_passenger_queue()

    def insert(self, p_passengers: List[AbstractPassenger]):
        """ Adds a new passenger to the database """
        if len(p_passengers) <= 0:
            return

        existing_passengers = self.database.get_passenger_queue_entries()

        for passenger in p_passengers:
            passenger.collect_log_guid(self.log.guid)
            self.log.append_text("Inserting passenger into queue: " + passenger.id_text)

            already_exists = False
            for existing_passenger in existing_passengers:
                if all([existing_passenger.passenger.source_system == passenger.source_system,
                        existing_passenger.passenger.external_id == passenger.external_id]):
                    already_exists = True
                    break

            if already_exists:
                self.log.append_entry(LogEntry(
                    p_message="Passenger already in queue, skipping",
                    p_type=MessageType.warning))
                continue

            client_passenger = self.database.client.get_client_passenger(passenger.passenger_module)
            pqs = QueueStatusFactory.get_passenger_queue_status(passenger, client_passenger)
            self.database.insert_passenger_queue(pqs)

    def get_deliverable_passengers(self, p_passenger_module: str) -> List[PassengerQueueStatus]:
        """ Returns passengers which can be pushed
        Passengers returned here are pulled, queued, processed and puller-notified,
        but not delivered
        """
        self.log.append_text("Reading deliverable passengers of type " + p_passenger_module)
        output = []

        candidates = self.database.get_passenger_queue_entries(
            p_passenger_module=p_passenger_module,
            p_pusher_status=QueueStatus.incomplete)

        for candidate in candidates:
            if candidate.all_processors_complete:
                output.append(candidate)

        return output

    def get_processable_passengers(self, p_passenger_module: str) -> List[PassengerQueueStatus]:
        """ Returns passengers which can be processed
        Passengers returned here are pulled, queued and puller-notified,
        but not processed
        """
        self.log.append_text("Reading processable passengers of type " + p_passenger_module)
        return self.database.get_passenger_queue_entries(
            p_passenger_module=p_passenger_module,
            p_processor_status=QueueStatus.incomplete)

    def get_puller_notifiable_passengers(self) -> List[PassengerQueueStatus]:
        """ Returns passengers which have been queued
        Passengers returned here are pulled, queued,
        but not puller-notified
        """
        self.log.append_text("Reading puller notifiable passengers")
        return self.database.get_passenger_queue_entries(p_puller_notified=False)

    def update_passenger_status(self, p_status: PassengerQueueStatus):
        """ Updates passenger status """
        p_status.passenger.collect_log_guid(self.log.guid)
        self.log.append_text("Updating passenger status for " + p_status.passenger.id_text)
        self.database.update_queue_status(p_status)



""" Abstract processor module """
from abc import ABC, abstractmethod
from typing import List
from databus.client.log import Log
from databus.pqueue.queue_status import PassengerQueueStatus


class AbstractProcessor(ABC): # pylint: disable=R0903
    """ Abstract processor class """
    def __init__(self, p_log: Log = None):
        self.log = p_log

    @abstractmethod
    def process(self, p_passengers: List[PassengerQueueStatus]):
        """ Processes the given passengers """



""" Generic JSON utility module """
from datetime import datetime


class JsonToolkit: # pylint: disable=R0903
    """ Generic JSON utility class """
    @staticmethod
    def convert_json_date_to_datetime(p_json_date: str) -> datetime:
        """ Converts JSON date string to PYTHON date time """
        return datetime.strptime(p_json_date, '%Y-%m-%dT%H:%M:%S.%f')



""" Demo passenger module """
from datetime import datetime
from uuid import UUID
from typing import List
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.passenger.attachment import Attachment


class DemoPassenger1(AbstractPassenger):
    """ Demo passenger class """
    def __init__(self, # pylint: disable=R0913
                 p_external_id: str = None,
                 p_internal_id: UUID = None,
                 p_source_system: str = None,
                 p_attachments: List[Attachment] = None,
                 p_puller_module: str = None,
                 p_pull_datetime: datetime = None,
                 p_passenger_module: str = None):

        if p_passenger_module is None:
            passenger_module = self.__module__
        else:
            passenger_module = p_passenger_module

        super().__init__(p_external_id=p_external_id,
                         p_internal_id=p_internal_id,
                         p_source_system=p_source_system,
                         p_attachments=p_attachments,
                         p_puller_module=p_puller_module,
                         p_pull_datetime=p_pull_datetime,
                         p_passenger_module=passenger_module)
        self.dataset = "Demo dataset"

    def hello_world(self):
        """ Dummy method """
        print("Demo passenger 1 says hello world!")
        print("My id is: " + str(self.internal_id))
        print("My dataset is: " + self.dataset)



""" E-Mail passenger module """
from datetime import datetime
import os
from typing import List
from uuid import UUID
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.passenger.attachment import Attachment


class Email(AbstractPassenger):
    """ E-Mail passenger class """
    _EXCEL_EXTENSIONS = [".xls", ".XLS", ".xlsx", ".XLSX"]
    _XML_EXTENSIONS = [".xml", ".XML"]

    def __init__(self, # pylint: disable=R0913
                 p_external_id: str = None,
                 p_internal_id: UUID = None,
                 p_source_system: str = None,
                 p_attachments: List[Attachment] = None,
                 p_puller_module: str = None,
                 p_pull_datetime: datetime = None,
                 p_passenger_module: str = None):

        if p_passenger_module is None:
            passenger_module = self.__module__
        else:
            passenger_module = p_passenger_module

        super().__init__(p_external_id=p_external_id,
                         p_internal_id=p_internal_id,
                         p_source_system=p_source_system,
                         p_attachments=p_attachments,
                         p_puller_module=p_puller_module,
                         p_pull_datetime=p_pull_datetime,
                         p_passenger_module=passenger_module)

    @property
    def excel_attachments(self) -> List[Attachment]:
        """ Returns a list of Excel attachments """
        return self.get_attachments_with_extensions(Email._EXCEL_EXTENSIONS)

    @property
    def xml_attachments(self) -> List[Attachment]:
        """ Returns a list of XML attachments """
        return self.get_attachments_with_extensions(Email._XML_EXTENSIONS)

    def get_attachments_with_extensions(self, p_extensions: List[str]) -> List[Attachment]:
        """ Returns attachments with given extensions """
        output = []
        for candidate in self.attachments:
            if "." not in candidate.name:
                continue
            file_name_parts = os.path.splitext(candidate.name)
            extension = file_name_parts[1].lower()
            if extension not in p_extensions:
                continue
            output.append(candidate)
        return output



""" Abstract puller module """
from abc import ABC, abstractmethod
from typing import List
from databus.client.log import Log
from databus.passenger.abstract_passenger import AbstractPassenger


class AbstractPuller(ABC):
    """ Abstract puller class """
    def __init__(self, p_log: Log = None):
        self.log = p_log

    @abstractmethod
    def notify_passengers_seated(self, p_seated_passengers: List[AbstractPassenger]):
        """ Called after a passenger is properly queued.
        You would typically write a code here to ensure that the passenger is not
        returned any more when the puller works again.
        """

    def peek(self) -> List[AbstractPassenger]:
        """ Peeks into the "inbox" of the source system
        and returns whatever awaits to be pulled.
        Normally, this method would simply pull and return
        whatever is in the inbox. However; if your source system
        needs to behave differently in peek / pull situations,
        you can override this method in your subclass.
        """
        return self.pull()

    @abstractmethod
    def pull(self) -> List[AbstractPassenger]:
        """ Pulls passengers from the source system """



""" Abstract driver factory module """
from abc import ABC, abstractmethod
from enum import Enum
from databus.driver.abstract_driver import AbstractDriver
from databus.processor.abstract_factory import AbstractProcessorFactory
from databus.pqueue.abstract_factory import AbstractQueueFactory
from databus.puller.abstract_factory import AbstractPullerFactory
from databus.pusher.abstract_factory import AbstractPusherFactory


class DriverCreationError(Exception):
    """ Driver creation exception """

    class ErrorCode(Enum):
        """ Driver creation error code """
        cant_create_instance: 1
        parameter_missing: 2

    def __init__(self, p_error_code: ErrorCode, p_module: str = ""):
        super().__init__()
        self.error_code = p_error_code

        if p_module is None:
            self.module = ""
        else:
            self.module = p_module

    @property
    def message(self) -> str:
        """ Error message as string """
        if self.error_code == DriverCreationError.ErrorCode.cant_create_instance:
            return "Can't create " + self.module + " driver instance"
        return "Driver creation error"


class AbstractDriverFactory(ABC): # pylint: disable=R0903
    """ Abstract driver factory class """
    @abstractmethod
    def create_driver(self, # pylint: disable=R0913
                      p_module: str,
                      p_queue_factory: AbstractQueueFactory,
                      p_processor_factory: AbstractProcessorFactory,
                      p_puller_factory: AbstractPullerFactory,
                      p_pusher_factory: AbstractPusherFactory
                      ) -> AbstractDriver:
        """ Driver creation """



""" Default pusher factory module """
from vibhaga.inspector import Inspector
from databus.client.log import Log
from databus.pusher.abstract_factory import AbstractPusherFactory, PusherCreationError
from databus.pusher.abstract_pusher import AbstractPusher


class PrimalPusherFactory(AbstractPusherFactory): # pylint: disable=R0903
    """ Default pusher factory class """

    def create_pusher(self, p_module: str, p_log: Log) -> AbstractPusher:
        """ Creates a new pusher object """
        if p_module == "" or p_module is None:
            raise PusherCreationError(PusherCreationError.ErrorCode.parameter_missing)

        candidates = Inspector.get_classes_in_module(p_module, exclude_classes=["AbstractPusher"])
        for candidate in candidates:
            try:
                obj_instance = candidate(p_log)
                if isinstance(obj_instance, AbstractPusher):
                    return obj_instance
            except Exception: # pylint: disable=W0703
                pass

        raise PusherCreationError(
            PusherCreationError.ErrorCode.cant_create_instance,
            p_module=p_module)



""" Abstract database factory """
from abc import ABC, abstractmethod
from enum import Enum
from typing import List
from databus.client.log import Log
from databus.database.abstract_database import AbstractDatabase
from databus.passenger.abstract_factory import AbstractPassengerFactory


class DatabaseCreationError(Exception):
    """ Central database creation exception class """

    class ErrorCode(Enum):
        """ Error code enum """
        cant_create_instance: 1
        parameter_missing: 2

    def __init__(self,
                 p_error_code: ErrorCode,
                 p_module: str = None,
                 p_client_id: str = None):
        super().__init__()
        self.error_code = p_error_code

        if p_module is None:
            self.module = ""
        else:
            self.module = p_module

        if p_client_id is None:
            self.client_id = ""
        else:
            self.client_id = p_client_id

    @property
    def message(self) -> str:
        """ Long text of the error """
        if self.error_code == DatabaseCreationError.ErrorCode.cant_create_instance:
            return "Can't create " + self.module + " instance for client " + self.client_id
        if self.error_code == DatabaseCreationError.ErrorCode.parameter_missing:
            return "Parameters missing, can't create database instance"
        return "Database creation error"


class AbstractDatabaseFactory(ABC): # pylint: disable=R0903
    """ Abstract database factory """
    @abstractmethod
    def create_database(self, # pylint: disable=R0913
                        p_module: str,
                        p_client_id: str,
                        p_log: Log,
                        p_passenger_factory: AbstractPassengerFactory,
                        p_arguments: dict
                        ) -> AbstractDatabase:
        """ Abstract method for database creation """

    @property
    @abstractmethod
    def database_modules(self) -> List[str]:
        """ Returns a list of database modules in the system """

    @property
    @abstractmethod
    def database_classes(self) -> List[str]:
        """ Returns a list of database classes in the system """



""" Module for data export operations """
from enum import Enum
from threading import Thread
from databus.client.log import Log, LogEntry, MessageType
from databus.database.abstract_database import AbstractDatabase
from databus.database.export.export_package import ExportPackage
from databus.database.export.export_request import ExportRequest


class ExportStatus(Enum):
    """ Defines an export status """
    IDLE = 1
    BUSY = 2


class Exporter:
    """ Executes an export request """

    def __init__(self, p_request: ExportRequest = None):
        self.request = p_request
        self._status = ExportStatus.IDLE

    @property
    def status(self) -> ExportStatus:
        """ Returns the status of the exporter """
        return self._status

    def execute(self, p_request: ExportRequest = None):
        """ Executes the export operation """
        self._prepare_before_execute(p_request)

        try:
            self._status = ExportStatus.BUSY
            log = Log()
            log.append_text("Starting data export")
            self.request.dispatcher.export_data_begin()
            source_db = self.request.dispatcher.get_client_database(self.request.client)
            package = Exporter._pull(source_db)
            Exporter._push(package, self.request.target_db)
        except Exception as error: # pylint: disable=W0703
            error_entry = LogEntry(
                str(error),
                p_type=MessageType.error)
            log.append_entry(error_entry)
        finally:
            self.request.dispatcher.export_data_end()
            log.append_text("Data export finished")
            try:
                self.request.target_db.insert_log(log)
            except Exception: # pylint: disable=W0703
                pass
            self._status = ExportStatus.IDLE
            print("Export complete")

    def execute_async(self, p_request: ExportRequest = None):
        """ Executes the export operation async """
        self._prepare_before_execute(p_request)
        Thread(target=self.execute(p_request), daemon=True).start()

    @staticmethod
    def _pull(p_database: AbstractDatabase) -> ExportPackage:
        output = ExportPackage(
            p_client=p_database.client,
            p_queues=p_database.get_passenger_queue_entries()
        )
        return output

    @staticmethod
    def _push(p_package: ExportPackage, p_database: AbstractDatabase):
        p_database.insert_client(p_package.client)

        for queue in p_package.queues:
            p_database.insert_passenger_queue(queue)

    def _prepare_before_execute(self, p_request: ExportRequest = None):
        if self._status == ExportStatus.BUSY:
            raise Exception("Exporter is busy, try later")
        if p_request is not None:
            self.request = p_request
        if self.request is None:
            raise Exception("No request provided")


class ExporterFactory: # pylint: disable=R0903
    """ Factory class for singleton Exporter instance """
    _SINGLETON = Exporter()

    @staticmethod
    def get_singleton() -> Exporter:
        """ Returns a singleton instance """
        return ExporterFactory._SINGLETON



""" Module for peek controllers """
from typing import List
from flask import render_template, request
from databus.client.client import Client
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher
from databus.passenger.abstract_passenger import AbstractPassenger, Attachment
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError


class PullerPeekResult: # pylint: disable=R0903
    """ Result of a pullers peek operation """
    def __init__(self, p_puller_module: str = "", p_passengers: List[AbstractPassenger] = None):
        self.puller_module = p_puller_module
        if p_passengers is None:
            self.passengers = []
        else:
            self.passengers = p_passengers


class ClientPeekResult: # pylint: disable=R0903
    """ Result of a clients peek operation """
    def __init__(self, p_client: Client = None, p_results: List[PullerPeekResult] = None):
        if p_client is None:
            self.client = Client()
        else:
            self.client = p_client

        if p_results is None:
            self.results = []
        else:
            self.results = p_results


class PullerPeek: # pylint: disable=R0903
    """ Class to peek into puller inboxes """
    def __init__(self, p_dispatcher: AbstractDispatcher):
        self._dispatcher = p_dispatcher

    def get_attachment(self,
                       p_client_id: str,
                       p_puller_module: str,
                       p_external_id: str,
                       p_attachment_name: str) -> Attachment:
        """ Returns an attachment """
        passenger = self.get_passenger(
            p_client_id=p_client_id,
            p_puller_module=p_puller_module,
            p_external_id=p_external_id)

        if passenger is None:
            return None

        for attachment in passenger.attachments:
            if attachment.name == p_attachment_name:
                return attachment

        return None

    def get_passenger(self,
                      p_client_id: str,
                      p_puller_module: str,
                      p_external_id: str) -> AbstractPassenger:
        """ Returns the requested passenger """
        peeks = self.peek()

        for peek in peeks:
            if peek.client.id != p_client_id:
                continue
            for client_result in peek.results:
                for passenger in client_result.passengers:
                    if passenger.puller_module == p_puller_module:
                        if passenger.external_id == p_external_id:
                            return passenger
        return None

    def peek(self, p_client_id: str = None) -> List[ClientPeekResult]:
        """ Peeks into puller inboxes """
        output = []
        driver = self._dispatcher.get_driver()

        for client in self._dispatcher.all_clients:
            if p_client_id is not None and p_client_id != client.id:
                continue
            client_peek = ClientPeekResult(p_client=client)
            for client_passenger in client.passengers:
                for puller_module in client_passenger.puller_modules:
                    puller_peek = PullerPeekResult(p_puller_module=puller_module)
                    new_passengers = driver.pull_passengers_from_module(puller_module)

                    for new_passenger in new_passengers:
                        puller_peek.passengers.append(new_passenger)
                    client_peek.results.append(puller_peek)
            output.append(client_peek)
        return output


class PeekAttachmentController(AbstractController):
    """ Peek attachment """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        puller = request.args.get("puller", 0, type=str)
        passenger = request.args.get("passenger", 0, type=str)
        file_name = request.args.get("file", 0, type=str)

        attachment = PullerPeek(self.dispatcher).get_attachment(
            p_client_id=self.requested_client_id,
            p_puller_module=puller,
            p_external_id=passenger,
            p_attachment_name=file_name)

        if attachment is None:
            return "File not found"

        return AbstractController._download_attachment(attachment)


class PeekController(AbstractController):
    """ Peek """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_root()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        peek = PullerPeek(self.dispatcher).peek(self.authenticated_client_id)
        return render_template("peek.html", peek=peek, alias=self.dispatcher.ticket.system_alias)



""" Module to build delete conditions """
from databus.database.sql_db.sql_database_arguments import SqlDatabaseArguments
from databus.database.sql_db.modifiable import Modifiable
from databus.database.sql_db.where_builder import WhereBuilder

class DeleteBuilder(Modifiable):
    """ Helper class to build update commands """
    def __init__(self, p_args: SqlDatabaseArguments, p_client_id: str = ""):
        super().__init__(p_args, p_client_id)
        self.where = WhereBuilder(p_client_id)

    @property
    def delete_command(self) -> str:
        """ Delete command """
        query = "DELETE FROM " + self.table
        query += self.where.where
        return query


    def clear(self):
        """ Reset """
        super().clear()
        try:
            self.where.clear()
        except Exception: # pylint: disable=W0703
            pass



""" Demo binary pusher module """
from os import path
from databus.pqueue.queue_status import PassengerQueueStatus, QueueStatus
from databus.pusher.abstract_pusher import AbstractPusher


class DemoBinaryPusher(AbstractPusher): # pylint: disable=R0903
    """ Demo binary pusher class """
    _BIN_FILE_NAME = "sample_binary_file.png"
    _MODULE_FILE_NAME = "demo_binary_pusher.py"

    def push(self, p_passenger: PassengerQueueStatus):
        """ Push demonstration """
        self.log.append_text("Pushing passenger " + p_passenger.passenger.id_text)
        DemoBinaryPusher._write_binary_file(p_passenger.passenger.attachments[0].binary_content)
        p_passenger.set_pusher_status(self.__module__, QueueStatus.complete)

    @staticmethod
    def _write_binary_file(p_bin: bytearray):
        """ Writes binary file to disk """
        full_path = path.abspath(__file__).replace(DemoBinaryPusher._MODULE_FILE_NAME,
                                                   DemoBinaryPusher._BIN_FILE_NAME)
        with open(full_path, "wb") as bin_file:
            bin_file.write(p_bin)



""" Abstract puller factory module """
from abc import ABC, abstractmethod
from enum import Enum
from databus.client.log import Log
from databus.puller.abstract_puller import AbstractPuller


class PullerCreationError(Exception):
    """ Puller creation exception class """

    class ErrorCode(Enum):
        """ Puller creation error code """
        cant_create_instance: 1
        parameter_missing: 2

    def __init__(self, p_error_code: ErrorCode, p_module: str = None):
        super().__init__()
        self.error_code = p_error_code

        if p_module is None:
            self.module = ""
        else:
            self.module = p_module

    @property
    def message(self) -> str:
        """ Puller creation error message """
        if self.error_code == PullerCreationError.ErrorCode.cant_create_instance:
            return "Can't create " + self.module + " puller instance"
        if self.error_code == PullerCreationError.ErrorCode.parameter_missing:
            return "Parameters missing, can't create puller instance"
        return "Puller creation error"


class AbstractPullerFactory(ABC): # pylint: disable=R0903
    """ Abstract puller factory class """
    @abstractmethod
    def create_puller(self, p_module: str, p_log: Log) -> AbstractPuller:
        """ Puller factory """



""" Abstract driver module """
from abc import ABC, abstractmethod
from typing import List
from databus.client.client_passenger import ClientPassenger
from databus.client.log import Log
from databus.database.abstract_database import AbstractDatabase
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.processor.abstract_factory import AbstractProcessorFactory
from databus.pqueue.abstract_factory import AbstractQueueFactory
from databus.pqueue.abstract_queue import AbstractQueue
from databus.puller.abstract_factory import AbstractPullerFactory
from databus.pusher.abstract_factory import AbstractPusherFactory


class BusTicket: # pylint: disable=R0903
    """ Driver constructor parameters """

    def __init__(self,
                 p_client_passenger: ClientPassenger = None,
                 p_log: Log = None,
                 p_database: AbstractDatabase = None):

        if p_client_passenger is None:
            self.client_passenger = ClientPassenger()
        else:
            self.client_passenger = p_client_passenger

        if p_log is None:
            self.log = Log()
        else:
            self.log = p_log

        self.database = p_database

    @property
    def client_id(self) -> str:
        """ ID of the client """
        return self.database.client.id


class AbstractDriver(ABC):
    """ Abstract driver class """

    def __init__(self,
                 p_queue_factory: AbstractQueueFactory,
                 p_processor_factory: AbstractProcessorFactory,
                 p_puller_factory: AbstractPullerFactory,
                 p_pusher_factory: AbstractPusherFactory):
        self.queue_factory = p_queue_factory
        self.processor_factory = p_processor_factory
        self.puller_factory = p_puller_factory
        self.pusher_factory = p_pusher_factory

    @property
    @abstractmethod
    def queue(self) -> AbstractQueue:
        """ Queue object """

    @abstractmethod
    def drive(self, p_bus_ticket: BusTicket):
        """ Carries passengers from source system to target system """

    def pull_passengers_from_module(self,
                                    p_puller_module: str,
                                    p_log: Log = None) -> List[AbstractPassenger]:
        """ Pulls new passengers from the given puller module """
        if p_log is None:
            log = Log()
        else:
            log = p_log
        puller_obj = self.puller_factory.create_puller(p_puller_module, log)
        return puller_obj.pull()



""" Utility """
from flask import session
from databus.client.user import Role

def get_authorization() -> dict:
    """ Returns authorization state """
    user_root = "client_id" in session and session["client_id"] == "root"
    user_admin = "user_role" in session and session["user_role"] == Role.ADMINISTRATOR.name
    return {
        "user_is_root": user_root,
        "user_is_admin": user_admin
    }

def user_is_admin() -> bool:
    """ Returns true if user is admin """
    return get_authorization()["user_is_admin"]



""" Abstract dispatcher module """
from abc import ABC, abstractmethod
from typing import List
from databus.client.client import Client
from databus.client.external_config import ExternalConfigFile, ExternalConfigFileManager
from databus.client.log import Log
from databus.database.abstract_database import AbstractDatabase
from databus.database.abstract_factory import AbstractDatabaseFactory
from databus.database.json_db.json_database_arguments import JsonDatabaseArguments
from databus.database.primal_factory import PrimalDatabaseFactory
from databus.dispatcher.observer import DispatcherObserver
from databus.driver.abstract_factory import AbstractDriverFactory
from databus.driver.abstract_driver import AbstractDriver
from databus.driver.primal_factory import PrimalDriverFactory
from databus.passenger.abstract_factory import AbstractPassengerFactory
from databus.passenger.primal_factory import PrimalPassengerFactory
from databus.pqueue.abstract_factory import AbstractQueueFactory
from databus.pqueue.primal_factory import PrimalQueueFactory
from databus.puller.abstract_factory import AbstractPullerFactory
from databus.puller.primal_factory import PrimalPullerFactory
from databus.pusher.abstract_factory import AbstractPusherFactory
from databus.pusher.primal_factory import PrimalPusherFactory
from databus.processor.abstract_factory import AbstractProcessorFactory
from databus.processor.primal_factory import PrimalProcessorFactory


class DispatcherTicket: # pylint: disable=R0902, R0903
    """ Factory parameters for dispatcher creation """
    _DEFAULT_DRIVER = "databus.driver.primal_driver"
    _DEFAULT_DATABASE = "databus.database.json_db.json_database"
    _DEFAULT_DATABASE_ARGS = {
        JsonDatabaseArguments.KEY_CLIENT_CONFIG: "config.json",
        JsonDatabaseArguments.KEY_CLIENT_DIR: "clients",
        JsonDatabaseArguments.KEY_DATABASE_DIR: "data|json_db",
        JsonDatabaseArguments.KEY_LOG_DIR: "log",
        JsonDatabaseArguments.KEY_LOG_EXTENSION: "txt",
        JsonDatabaseArguments.KEY_QUEUE_ATTACHMENT_DIR: "attachments",
        JsonDatabaseArguments.KEY_QUEUE_DIR: "pqueue",
        JsonDatabaseArguments.KEY_QUEUE_PASSENGER: "passenger.json"
    }

    def __init__(self,
                 p_database_factory: AbstractDatabaseFactory = None,
                 p_driver_factory: AbstractDriverFactory = None,
                 p_passenger_factory: AbstractPassengerFactory = None,
                 p_queue_factory: AbstractQueueFactory = None,
                 p_puller_factory: AbstractPullerFactory = None,
                 p_processor_factory: AbstractProcessorFactory = None,
                 p_pusher_factory: AbstractPusherFactory = None,
                 p_database_module: str = None,
                 p_database_arguments: dict = None,
                 p_driver_module: str = None,
                 p_dispatcher_observer: DispatcherObserver = None,
                 p_run_web_server: bool = True,
                 p_web_server_port: int = 5000,
                 p_external_config_files: List[ExternalConfigFile] = None,
                 p_system_alias: str = None
                 ): # pylint: disable=R0912, R0913, R0915, R0914

        if p_database_factory is None:
            self.database_factory = PrimalDatabaseFactory()
        else:
            self.database_factory = p_database_factory

        if p_driver_factory is None:
            self.driver_factory = PrimalDriverFactory()
        else:
            self.driver_factory = p_driver_factory

        if p_passenger_factory is None:
            self.passenger_factory = PrimalPassengerFactory()
        else:
            self.passenger_factory = p_passenger_factory

        if p_queue_factory is None:
            self.queue_factory = PrimalQueueFactory()
        else:
            self.queue_factory = p_queue_factory

        if p_puller_factory is None:
            self.puller_factory = PrimalPullerFactory()
        else:
            self.puller_factory = p_puller_factory

        if p_processor_factory is None:
            self.processor_factory = PrimalProcessorFactory()
        else:
            self.processor_factory = p_processor_factory

        if p_pusher_factory is None:
            self.pusher_factory = PrimalPusherFactory()
        else:
            self.pusher_factory = p_pusher_factory

        if p_database_module is None:
            self.database_module = DispatcherTicket._DEFAULT_DATABASE
        else:
            self.database_module = p_database_module

        if p_database_arguments is None:
            self.database_arguments = DispatcherTicket._DEFAULT_DATABASE_ARGS
        else:
            self.database_arguments = p_database_arguments

        if p_driver_module is None:
            self.driver_module = DispatcherTicket._DEFAULT_DRIVER
        else:
            self.driver_module = p_driver_module

        if p_external_config_files is None:
            self.external_config_files = []
        else:
            self.external_config_files = p_external_config_files

        if p_system_alias is None or p_system_alias == "":
            self.system_alias = "Databus"
        else:
            self.system_alias = p_system_alias

        self.dispatcher_observer = p_dispatcher_observer
        self.run_web_server = p_run_web_server
        self.web_server_port = p_web_server_port


class AbstractDispatcher(ABC): # pylint: disable=R0903
    """ Abstract dispatcher class """
    def __init__(self, p_ticket: DispatcherTicket = None):
        self.ticket = p_ticket
        self.external_config_file_manager = ExternalConfigFileManager()
        self.external_config_file_manager.add_files(self.ticket.external_config_files)

    @property
    def all_clients(self) -> List[Client]:
        """ Returns a list of clients from the database within the ticket """
        dummy_db = self.ticket.database_factory.create_database(
            p_log=Log(),
            p_module=self.ticket.database_module,
            p_client_id=None,
            p_passenger_factory=self.ticket.passenger_factory,
            p_arguments=self.ticket.database_arguments)

        return dummy_db.get_clients()

    @abstractmethod
    def expedite_client_passenger(self, p_client_id: str, p_passenger_module: str):
        """ Prioritizes the passenger in the next cycle """

    @abstractmethod
    def export_data_begin(self):
        """ Indicates that data export is starting """

    @abstractmethod
    def export_data_end(self):
        """ Indicates that data export is ending """

    def get_client_database(self, p_client_id: str, p_log: Log = None) -> AbstractDatabase:
        """ Returns a database instance for the given client """
        if p_log is None:
            log = Log()
        else:
            log = p_log

        return self.ticket.database_factory.create_database( # pylint: disable=C0103
            p_passenger_factory=self.ticket.passenger_factory,
            p_client_id=p_client_id,
            p_module=self.ticket.database_module,
            p_log=log,
            p_arguments=self.ticket.database_arguments)

    def get_driver(self) -> AbstractDriver:
        """ Returns a new driver instance """
        driver = self.ticket.driver_factory.create_driver(
            self.ticket.driver_module,
            self.ticket.queue_factory,
            self.ticket.processor_factory,
            self.ticket.puller_factory,
            self.ticket.pusher_factory)
        return driver

    @abstractmethod
    def start(self):
        """ Activates dispatcher timer """

    @abstractmethod
    def request_shutdown(self):
        """ Dispatcher shutdown """



""" Demo processor module """
from typing import List
from databus.pqueue.queue_status import PassengerQueueStatus, QueueStatus
from databus.processor.abstract_processor import AbstractProcessor



class DemoProcessor1(AbstractProcessor): # pylint: disable=R0903
    """ Demo processor class """

    def process(self, p_passengers: List[PassengerQueueStatus]):
        """ Demo process """
        if len(p_passengers) > 1:
            self.log.append_text("Marking last passenger as processed & pushed")
            last_passenger = p_passengers[len(p_passengers)-1]
            last_passenger.set_all_processor_statuses(QueueStatus.complete)
            last_passenger.set_all_pusher_statuses(QueueStatus.complete)
            self.log.append_text("Remaining passengers: " + str(len(p_passengers)))
        else:
            return



""" Path builder module """
from os import path, scandir
from typing import List
from databus import get_root_path
from databus.database.json_db.json_database_arguments import JsonDatabaseArguments

class JsonPathBuilder:
    """ JSON path builder class """

    @staticmethod
    def get_client_root_path(p_args: JsonDatabaseArguments) -> str:
        """ Builds the concrete root path of client directories """
        databus_root = get_root_path()
        return path.join(databus_root, p_args.database_dir, p_args.client_dir)

    def __init__(self, p_client_id: str, p_args: JsonDatabaseArguments):
        self._client_id = p_client_id
        self._args = p_args

    @property
    def client_root_path(self) -> str:
        """ Returns root path of client """
        return JsonPathBuilder.get_client_root_path(self._args)

    @property
    def client_dir_path(self) -> str:
        """ Builds the concrete path for client directory """
        client_root_path = self.client_root_path
        return path.join(client_root_path, self._client_id)

    @property
    def config_file_path(self) -> str:
        """ Returns the config file path """
        return path.join(
            self.client_dir_path,
            self._args.client_config)

    @property
    def log_root_path(self) -> str:
        """ Builds log file root path """
        return path.join(self.client_dir_path,
                         self._args.log_dir)

    @property
    def passenger_directories(self) -> List[str]:
        """ Returns all passenger directories of client """
        return [f.name for f in scandir(self.queue_root_path) if f.is_dir()]

    @property
    def queue_root_path(self) -> str:
        """ Returns the root path of the queue dir """
        return path.join(
            self.client_dir_path,
            self._args.queue_dir)

    def get_attachment_directory_path(self, p_internal_id: str) -> str:
        """ Returns the attachment directory path """
        return path.join(
            self.queue_root_path,
            p_internal_id,
            self._args.queue_attachment_dir)

    def get_attachment_file_path(self, p_internal_id: str, p_file_name: str) -> str:
        """ Returns the path of the given attachment file """
        return path.join(
            self.get_attachment_directory_path(p_internal_id),
            p_file_name)

    def get_passenger_directory_path(self, p_internal_id: str) -> str:
        """ Returns the path of the given passenger directory """
        return path.join(
            self.queue_root_path,
            p_internal_id)

    def get_passenger_file_path(self, p_internal_id: str) -> str:
        """ Returns passenger file path """
        return path.join(
            self.queue_root_path,
            p_internal_id,
            self._args.queue_passenger)



""" Demo processor module """
from typing import List
from databus.client.log import LogEntry, MessageType
from databus.pqueue.queue_status import PassengerQueueStatus
from databus.processor.abstract_processor import AbstractProcessor


class DemoProcessor1(AbstractProcessor): # pylint: disable=R0903
    """ Demo processor class """

    def process(self, p_passengers: List[PassengerQueueStatus]):
        """ Demo process """

        self.log.entries.append(LogEntry(p_message="Doing nothing",
                                         p_type=MessageType.warning))



""" Default passenger factory module """
from vibhaga.inspector import Inspector
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.passenger.abstract_factory import AbstractPassengerFactory, PassengerCreationError


class PrimalPassengerFactory(AbstractPassengerFactory): # pylint: disable=R0903
    """ Default passenger factory class """

    def create_passenger(self, p_module: str) -> AbstractPassenger:
        """ Default passenger factory """
        if p_module == "" or p_module is None:
            raise PassengerCreationError(PassengerCreationError.ErrorCode.parameter_missing)

        candidates = Inspector.get_classes_in_module(p_module, exclude_classes=["AbstractPassenger"]) # pylint: disable=C0301
        for candidate in candidates:
            try:
                obj_instance = candidate(p_passenger_module=p_module)
                if obj_instance.__module__ != p_module:
                    continue
                if isinstance(obj_instance, AbstractPassenger):
                    return obj_instance
            except Exception: # pylint: disable=W0703
                pass

        raise PassengerCreationError(
            PassengerCreationError.ErrorCode.cant_create_instance,
            p_module=p_module)



""" Module for passenger pages """
from flask import render_template, redirect, request, url_for
from databus.client.client import Client
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError


class PassengerExpediteController(AbstractController):
    """ Passenger expedite """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        passenger = request.args.get("passenger", 0, type=str)
        self.dispatcher.expedite_client_passenger(self.requested_client_id, passenger)
        url = url_for("_passenger_list")
        url += "?expedited=true"
        url += AbstractController._get_cache_buster()
        return redirect(url, code=302)


class PassengerListController(AbstractController):
    """ Passenger list page """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        if self.authenticated_client_id == Client.ROOT:
            clients = self.dispatcher.all_clients
        else:
            clients = [self.authenticated_client_database.client]

        try:
            expedited = request.args.get("expedited", 0, type=str) == "true"
        except Exception: # pylint: disable=W0703
            expedited = False

        return render_template(
            "passenger_list.html",
            clients=clients,
            expedited=expedited,
            alias=self.dispatcher.ticket.system_alias)



""" Default dispatcher factory module """
from vibhaga.inspector import Inspector
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher, DispatcherTicket
from databus.dispatcher.abstract_factory import \
    AbstractDispatcherFactory, DispatcherCreationError


class PrimalDispatcherFactory(AbstractDispatcherFactory): # pylint: disable=R0903
    """ Default dispatcher factory class """
    _DEFAULT_DISPATCHER = "databus.dispatcher.primal_dispatcher"

    def create_dispatcher(self,
                          p_module: str = None,
                          p_ticket: DispatcherTicket = None) -> AbstractDispatcher:
        """ Dispatcher factory """

        if p_module == "" or p_module is None:
            dispatcher_module = PrimalDispatcherFactory._DEFAULT_DISPATCHER
        else:
            dispatcher_module = p_module

        if p_ticket is None:
            dispatcher_ticket = DispatcherTicket()
        else:
            dispatcher_ticket = p_ticket

        candidates = Inspector.get_classes_in_module(
            dispatcher_module,
            exclude_classes=["AbstractDispatcher"])

        for candidate in candidates:
            try:
                obj_instance = candidate(dispatcher_ticket)
                if isinstance(obj_instance, AbstractDispatcher):
                    return obj_instance
            except Exception: # pylint: disable=W0703
                pass

        raise DispatcherCreationError(
            DispatcherCreationError.ErrorCode.cant_create_instance,
            p_module=dispatcher_module)



""" Module for home type of pages """
from flask import render_template
import databus
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError


class AboutController(AbstractController):
    """ About page """

    def execute(self):
        """ Builds and returns the page """

        return render_template(
            "about.html",
            version=databus.__version__,
            author=databus.AUTHOR,
            email=databus.EMAIL,
            description=databus.DESCRIPTION,
            python_version=databus.PYTHON_VERSION,
            dispatcher=self.dispatcher,
            alias=self.dispatcher.ticket.system_alias)


class HomeController(AbstractController):
    """ Home page """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate()
        except AuthenticationError as authentication_error:
            return authentication_error.output

        return render_template("home.html", alias=self.dispatcher.ticket.system_alias)



""" Shutdown web module """
from flask import redirect, url_for, render_template
from databus.client.client import Client
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError

class ShutdownController(AbstractController):
    """ Shutdown controller """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        if self.authenticated_client_id != Client.ROOT:
            return redirect(url_for("_login"), code=302)

        return render_template("shutdown.html", alias=self.dispatcher.ticket.system_alias)

class ShutdownExeController(AbstractController):
    """ Shutdown execution """

    def execute(self):
        """ Shutdown execute """
        try:
            self._authenticate(must_be_admin=True)
        except AuthenticationError:
            return "Error shutting down, cant authenticate"

        if self.authenticated_client_id != Client.ROOT:
            return "Error shutting down, no permission"

        self.dispatcher.request_shutdown()
        return "Shutdown started"



""" Module for user pages """
from flask import render_template, redirect, request, url_for
from databus.client.client import Client
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError

class UserTokenRevokeController(AbstractController):
    """ User token revoke """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        client_db = self.requested_client_database
        username = request.args.get("user", 0, type=str)

        for user in client_db.client.users:
            if user.credential.username == username:
                user.credential.token = ""
                client_db.update_user_credential(user.credential)
                break

        url = url_for("_user_list")
        url += "?" + AbstractController._get_cache_buster()
        return redirect(url, code=302)

class UserListController(AbstractController):
    """ User list page """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        if self.authenticated_client_id == Client.ROOT:
            clients = self.dispatcher.all_clients
        else:
            clients = [self.authenticated_client_database.client]

        return render_template(
            "user_list.html",
            clients=clients,
            alias=self.dispatcher.ticket.system_alias)



""" Path builder module
This module includes helper methods to build
table & field paths
"""
from databus.database.sql_db.sql_database_arguments import SqlDatabaseArguments


class PathBuilder:
    """ Path builder class for table & field paths """
    def __init__(self, p_args: SqlDatabaseArguments):
        self._args = p_args

    def get_field_path(self, p_table: str, p_field: str) -> str:
        """ Returns the full path of the table field
        DATABASE.SCHEMA.TABLE.FIELD
        """
        return self.get_table_path(p_table) + "." + p_field

    def get_table_path(self, p_table: str) -> str:
        """ Returns the full path of the table
        DATABASE.SCHEMA.TABLE
        """
        return self._args.database + "." + self._args.schema + "." + p_table



""" Demo puller module """
from typing import List
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.passenger.demo.demo_passenger_2 import DemoPassenger2
from databus.puller.abstract_puller import AbstractPuller


class DemoPuller2(AbstractPuller):
    """ Demo puller class """

    def notify_passengers_seated(self, p_seated_passengers: List[AbstractPassenger]):
        """ Fake operation """
        for seated_passenger in p_seated_passengers:
            self.log.append_text(
                "Demo puller 2 notified about seated passenger " + seated_passenger.id_text)

    def pull(self) -> List[DemoPassenger2]:
        """ Fake operation """
        output = []

        passenger1 = DemoPassenger2()
        passenger1.external_id = "ID_2_1"
        passenger1.dataset = "Puller 2 pulled first DemoPassenger2"
        passenger1.source_system = "DEMO_SYSTEM"
        passenger1.puller_module = self.__module__
        output.append(passenger1)
        self.log.append_text("Got passenger " + passenger1.id_text)

        passenger2 = DemoPassenger2()
        passenger2.external_id = "ID_2_2"
        passenger2.dataset = "Puller 2 pulled second DemoPassenger2"
        passenger2.source_system = "DEMO_SYSTEM"
        passenger2.puller_module = self.__module__
        output.append(passenger2)
        self.log.append_text("Got passenger " + passenger2.id_text)

        return output



""" Module for a database table """
from typing import List
from databus.database.sql_db.query_helper import QueryHelper


class Table: # pylint: disable=R0903
    """ Database table class """
    _key_fields: dict = {}

    def __init__(self, p_query_helper: QueryHelper, p_name: str):
        self._query_helper = p_query_helper
        self._name = p_name

    @property
    def key_fields(self) -> List[str]:
        """ Returns a list of key fields """
        if self._name not in Table._key_fields:
            Table._key_fields[self._name] = []

            primary_keys = self._query_helper.select_all_literal_where(
                "primary_keys",
                p_literal_where=" WHERE table_name = '" + self._name + "'")

            for primary_key in primary_keys:
                if primary_key["column_name"] not in Table._key_fields[self._name]:
                    Table._key_fields[self._name].append(primary_key["column_name"])

        return Table._key_fields[self._name]



""" Module for connecting to SQL Server over PYODBC """
from typing import List
import pyodbc
from databus.database.sql_db.sql_database_arguments import SqlDatabaseArguments
from databus.database.sql_db.driver.abstract_driver import SqlServerDriver


class PyodbcDriver(SqlServerDriver):
    """ PYODBC driver for SQL Server """

    def __init__(self):
        super().__init__()
        self._connection = None
        self._cursor = None

    def commit(self):
        """ Commits the transactions """
        self._connection.commit()

    def connect(self, p_args: SqlDatabaseArguments):
        """ Opens a new connection to the database """
        self._connection = pyodbc.connect( # pylint: disable=I1101
            "DRIVER={ODBC Driver 17 for SQL Server};"
            "SERVER=" + p_args.server + ";"
            "DATABASE=" + p_args.database + ";"
            "UID=" + p_args.username + ";"
            "PWD=" + p_args.password + ";")

        self._cursor = self._connection.cursor()

    def execute_sql(self, p_query: str):
        """ Executes the given SQL
        Also commits if self.autocommit is true
        """
        self._cursor.execute(p_query)
        if self.autocommit:
            self.commit()

    def execute_stored_procedure(self, p_sql: str, p_values):
        """ Executes a stored procedure """
        self._cursor.execute(p_sql, (p_values))

    def rollback(self):
        """ Rollbacks the transactions """
        self._connection.rollback()

    def select(self, p_query: str) -> List[dict]:
        """ Selects data from the database """
        output = []
        self._cursor.execute(p_query)
        results = self._cursor.fetchall()
        for result in results:
            output_item = {}
            index = 0
            for description in result.cursor_description:
                output_item[description[0]] = result[index]
                index += 1
            output.append(output_item)
        return output



""" Abstract pusher module """
from abc import ABC, abstractmethod
from databus.client.log import Log
from databus.pqueue.queue_status import PassengerQueueStatus


class AbstractPusher(ABC): # pylint: disable=R0903
    """ Abstract pusher class """
    def __init__(self, p_log: Log = None):
        self.log = p_log

    @abstractmethod
    def push(self, p_passenger: PassengerQueueStatus):
        """ Sends passenger to the target system """



""" Module for web users """
import uuid
from enum import Enum


class Credential: # pylint: disable=R0903
    """ Class defining a user credential """
    def __init__(self, username: str = "Guest", password: str = "", token: str = ""):
        self.username = username
        self.password = password
        self.token = token

    def generate_token(self):
        """ Generates and assigns a new token """
        self.token = str(uuid.uuid1())


class Role(Enum):
    """ Defines the role of the user within the system """
    UNDEFINED = 0
    OPERATOR = 1
    ADMINISTRATOR = 2


def str_to_role(role: str) -> Role:
    """ Converts a text based role value to role enum """
    up_role = role.upper()
    for role_enum in Role:
        if role_enum.name == up_role:
            return role_enum
    return Role.UNDEFINED


class User: # pylint: disable=R0903
    """ Class defining a web user """
    def __init__(self, credential: Credential = None, role: Role = Role.OPERATOR):
        if credential is None:
            self.credential = Credential()
        else:
            self.credential = credential

        if role == Role.UNDEFINED:
            self.role = Role.OPERATOR
        else:
            self.role = role

    def authenticate(self, credential: Credential) -> bool:
        """ Checks if the user & password matches """
        if credential.username != self.credential.username:
            return False
        if credential.password != "" and credential.password == self.credential.password:
            return True
        if credential.token != "" and credential.token == self.credential.token:
            return True
        return False



""" Abstract queue module """
from abc import ABC, abstractmethod
from typing import List
from datetime import datetime
from databus.client.log import Log
from databus.database.abstract_database import AbstractDatabase
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.pqueue.queue_status import PassengerQueueStatus


class AbstractQueue(ABC):
    """ Abstract queue class """

    def __init__(self, p_database: AbstractDatabase, p_log: Log):
        self.database = p_database
        self.log = p_log

    @abstractmethod
    def delete_completed_passengers(self, p_passenger_module: str, p_pulled_before: datetime):
        """ Deletes completed passengers from the database,
        which were pulled before p_pulled_before and queued, processed, puller-notified, pushed
        """

    @abstractmethod
    def erase(self):
        """ Deletes all passengers from the database """

    @abstractmethod
    def insert(self, p_passengers: List[AbstractPassenger]):
        """ Adds a new passenger to the database """

    @abstractmethod
    def get_deliverable_passengers(self, p_passenger_module: str) -> List[PassengerQueueStatus]:
        """ Returns passengers which can be pushed
        Passengers returned here are pulled, queued, processed and puller-notified,
        but not delivered
        """

    @abstractmethod
    def get_processable_passengers(self, p_passenger_module: str) -> List[PassengerQueueStatus]:
        """ Returns passengers which can be processed
        Passengers returned here are pulled, queued and puller-notified,
        but not processed
        """

    @abstractmethod
    def get_puller_notifiable_passengers(self) -> List[PassengerQueueStatus]:
        """ Returns passengers which have been queued
        Passengers returned here are pulled, queued,
        but not puller-notified
        """

    @abstractmethod
    def update_passenger_status(self, p_status: PassengerQueueStatus):
        """ Updates passenger status """



""" Default queue factory module """
from vibhaga.inspector import Inspector
from databus.client.log import Log
from databus.database.abstract_database import AbstractDatabase
from databus.pqueue.abstract_factory import AbstractQueueFactory, QueueCreationError
from databus.pqueue.abstract_queue import AbstractQueue


class PrimalQueueFactory(AbstractQueueFactory): # pylint: disable=R0903
    """ Default queue factory class """

    def create_queue(self,
                     p_module: str,
                     p_database: AbstractDatabase,
                     p_log: Log) -> AbstractQueue:
        """ Queue factory """

        if p_module == "" or p_module is None or p_database is None or p_log is None:
            raise QueueCreationError(
                QueueCreationError.ErrorCode.parameter_missing,
                p_module=str(p_module))

        candidates = Inspector.get_classes_in_module(p_module, exclude_classes=["AbstractQueue"])
        for candidate in candidates:
            try:
                obj_instance = candidate(p_database, p_log)
                if isinstance(obj_instance, AbstractQueue):
                    return obj_instance
            except Exception: # pylint: disable=W0703
                continue

        raise QueueCreationError(
            QueueCreationError.ErrorCode.cant_create_instance,
            p_module=p_module)



""" Default puller factory module """
from vibhaga.inspector import Inspector
from databus.client.log import Log
from databus.puller.abstract_factory import AbstractPullerFactory, PullerCreationError
from databus.puller.abstract_puller import AbstractPuller


class PrimalPullerFactory(AbstractPullerFactory):  # pylint: disable=R0903
    """ Default puller factory class """

    def create_puller(self, p_module: str, p_log: Log) -> AbstractPuller:
        """ Puller factory """
        if p_module == "" or p_module is None:
            raise PullerCreationError(PullerCreationError.ErrorCode.parameter_missing)

        candidates = Inspector.get_classes_in_module(p_module, exclude_classes=["AbstractPuller"])

        for candidate in candidates:
            try:
                obj_instance = candidate(p_log)
                if isinstance(obj_instance, AbstractPuller) and obj_instance.__module__ == p_module:
                    return obj_instance
            except Exception: # pylint: disable=W0703
                continue

        raise PullerCreationError(
            PullerCreationError.ErrorCode.cant_create_instance,
            p_module=p_module)



""" Module for an export package """
from typing import List
from databus.client.client import Client
from databus.pqueue.queue_status import PassengerQueueStatus


class ExportPackage: # pylint: disable=R0903
    """ Defines an export package """

    def __init__(self,
                 p_client: Client = None,
                 p_queues: List[PassengerQueueStatus] = None):
        self.client = p_client

        if p_queues is None:
            self.queues = []
        else:
            self.queues = p_queues



""" Json database argument module """
from enum import Enum
from os import path


class JsonDatabaseArgumentError(Exception):
    """ Json database argument exception """

    class ErrorCode(Enum):
        """ Error code """
        invalid_argument: 1

    def __init__(self, p_error_code: ErrorCode, p_argument: str = ""):
        super().__init__()
        self.error_code = p_error_code

        if p_argument is None:
            self.argument = ""
        else:
            self.argument = p_argument

    @property
    def message(self) -> str:
        """ Error message as string """
        if self.error_code == JsonDatabaseArgumentError.ErrorCode.invalid_argument:
            return "Invalid JsonDatabase argument: " + self.argument
        return "JsonDatabase argument error"


class JsonDatabaseArguments: # pylint: disable=R0902, R0903
    """ Json database argument class """
    KEY_CLIENT_CONFIG = "client_config"
    KEY_CLIENT_DIR = "client_dir"
    KEY_DATABASE_DIR = "database_dir"
    KEY_LOG_DIR = "log_dir"
    KEY_LOG_EXTENSION = "log_extension"
    KEY_QUEUE_ATTACHMENT_DIR = "queue_attachment_dir"
    KEY_QUEUE_DIR = "queue_dir"
    KEY_QUEUE_PASSENGER = "queue_passenger"

    TEMPLATE = {
        KEY_CLIENT_CONFIG: "config.json",
        KEY_CLIENT_DIR: "clients",
        KEY_DATABASE_DIR: "databus|database|json_db",
        KEY_LOG_DIR: "log",
        KEY_LOG_EXTENSION: "txt",
        KEY_QUEUE_ATTACHMENT_DIR: "attachments",
        KEY_QUEUE_DIR: "pqueue",
        KEY_QUEUE_PASSENGER: "passenger.json"
    }

    def __init__(self, p_arguments: dict):
        self.client_config = JsonDatabaseArguments.TEMPLATE[JsonDatabaseArguments.KEY_CLIENT_CONFIG]
        self.client_dir = JsonDatabaseArguments.TEMPLATE[JsonDatabaseArguments.KEY_CLIENT_DIR]
        self._database_dir = JsonDatabaseArguments.TEMPLATE[JsonDatabaseArguments.KEY_DATABASE_DIR]
        self.log_dir = JsonDatabaseArguments.TEMPLATE[JsonDatabaseArguments.KEY_LOG_DIR]
        self.log_extension = JsonDatabaseArguments.TEMPLATE[JsonDatabaseArguments.KEY_LOG_EXTENSION]
        self.queue_attachment_dir = JsonDatabaseArguments.TEMPLATE[JsonDatabaseArguments.KEY_QUEUE_ATTACHMENT_DIR] # pylint: disable=C0301
        self.queue_dir = JsonDatabaseArguments.TEMPLATE[JsonDatabaseArguments.KEY_QUEUE_DIR]
        self.queue_passenger = JsonDatabaseArguments.TEMPLATE[JsonDatabaseArguments.KEY_QUEUE_PASSENGER] # pylint: disable=C0301

        for key in p_arguments:
            if key == JsonDatabaseArguments.KEY_CLIENT_CONFIG:
                self.client_config = p_arguments[key]
            elif key == JsonDatabaseArguments.KEY_CLIENT_DIR:
                self.client_dir = p_arguments[key]
            elif key == JsonDatabaseArguments.KEY_DATABASE_DIR:
                self._database_dir = p_arguments[key]
            elif key == JsonDatabaseArguments.KEY_LOG_DIR:
                self.log_dir = p_arguments[key]
            elif key == JsonDatabaseArguments.KEY_LOG_EXTENSION:
                self.log_extension = p_arguments[key]
            elif key == JsonDatabaseArguments.KEY_QUEUE_ATTACHMENT_DIR:
                self.queue_attachment_dir = p_arguments[key]
            elif key == JsonDatabaseArguments.KEY_QUEUE_DIR:
                self.queue_dir = p_arguments[key]
            elif key == JsonDatabaseArguments.KEY_QUEUE_PASSENGER:
                self.queue_passenger = p_arguments[key]
            else:
                raise JsonDatabaseArgumentError(
                    JsonDatabaseArgumentError.ErrorCode.invalid_argument,
                    p_argument=key)

    @property
    def database_dir(self) -> str:
        """ Database directory """
        if "|" in self._database_dir:
            output = ""
            fragments = self._database_dir.split("|")
            for fragment in fragments:
                output = path.join(output, fragment)
            return output
        return self._database_dir



""" Module for an abstract SQL Server driver """
from abc import ABC, abstractmethod
from typing import List
from databus.database.sql_db.sql_database_arguments import SqlDatabaseArguments


class SqlServerDriver(ABC):
    """ Abstract SQL Server driver class """

    def __init__(self):
        self.autocommit = True

    @abstractmethod
    def commit(self):
        """ Commits the transactions """

    @abstractmethod
    def connect(self, p_args: SqlDatabaseArguments):
        """ Opens a new connection to the database """

    @abstractmethod
    def execute_stored_procedure(self, p_sql: str, p_values):
        """ Executes a stored procedure """

    @abstractmethod
    def execute_sql(self, p_query: str):
        """ Executes the given SQL
        Also commits if self.autocommit is true
        """

    @abstractmethod
    def rollback(self):
        """ Rollbacks the transactions """

    @abstractmethod
    def select(self, p_query: str) -> List[dict]:
        """ Selects data from the database """



""" Abstract queue factory module """
from abc import abstractmethod
from enum import Enum
from databus.client.log import Log
from databus.database.abstract_database import AbstractDatabase
from databus.pqueue.abstract_queue import AbstractQueue


class QueueCreationError(Exception):
    """ Queue creation exception """

    class ErrorCode(Enum):
        """ Queue creation error code """
        cant_create_instance: 1
        parameter_missing: 2

    def __init__(self,
                 p_error_code: ErrorCode,
                 p_module: str = ""):

        super().__init__()
        self.error_code = p_error_code

        if p_module is None:
            self.module = ""
        else:
            self.module = p_module

    @property
    def message(self) -> str:
        """ Error message as text """
        if self.error_code == QueueCreationError.ErrorCode.cant_create_instance:
            return "Can't create " + self.module + " queue instance"
        if self.error_code == QueueCreationError.ErrorCode.parameter_missing:
            return "Parameters missing, can't create database instance"
        return "Database creation error"


class AbstractQueueFactory: # pylint: disable=R0903
    """ Abstract queue factory class """
    @abstractmethod
    def create_queue(self,
                     p_module: str,
                     p_database: AbstractDatabase,
                     p_log: Log) -> AbstractQueue:
        """ Queue factory """



""" Module for customizing pages """
from typing import List
from flask import render_template, redirect, request, url_for
from databus.client.client import Client
from databus.dispatcher.abstract_dispatcher import AbstractDispatcher
from databus.web.controller.abstract_controller import AbstractController, AuthenticationError


_CLIENTS_DB_NODE = "__clients__"
_DATABUS_DB_NODE = "__databus__"

class CustomizingNode:  # pylint: disable=R0903
    """ Defines a customizing node """
    def __init__(self, p_name: str = "", p_content: str = ""):
        self.name = p_name
        self.content = p_content


class ClientCustomizing: # pylint: disable=R0903
    """ Defines a client and its customizing files """
    def __init__(self, p_client: Client = None, p_customs: List[CustomizingNode] = None):
        if p_client is None:
            self.client = Client()
        else:
            self.client = p_client

        if p_customs is None:
            self.nodes = []
        else:
            self.nodes = p_customs


class ClientCustomizingReader: # pylint: disable=R0903
    """ Class dealing with client customizing files """
    def __init__(self, p_dispatcher: AbstractDispatcher):
        self._dispatcher = p_dispatcher

    def get_client_customizing_entry(self, p_client_id: str, p_entry_name: str) -> ClientCustomizing: # pylint: disable=C0301
        """ Returns a single node """
        all_entries = self.get_client_customizing_list(p_client_id=p_client_id)
        if all_entries is None or len(all_entries) < 0:
            return None
        for entry in all_entries[0].nodes:
            if entry.name == p_entry_name:
                return entry
        return None

    def get_client_customizing_list(self, p_client_id: str = None):
        """ Returns a list of client - customizing files """
        output = []

        for client in self._dispatcher.all_clients:
            if p_client_id is not None and client.id != p_client_id:
                continue
            client_nodes = []

            client_database = self._dispatcher.get_client_database(client.id)
            db_node = CustomizingNode(_DATABUS_DB_NODE, client_database.customizing)
            client_nodes.append(db_node)

            if client.id == Client.ROOT:
                clients_node = CustomizingNode(_CLIENTS_DB_NODE, client_database.client_master_data)
                client_nodes.append(clients_node)

            external_files = self._dispatcher.external_config_file_manager.get_files_of_client(client.id) # pylint: disable=C0301
            for external_file in external_files:
                external_node = CustomizingNode(external_file.file_id, external_file.file_content)
                client_nodes.append(external_node)

            output.append(ClientCustomizing(client, client_nodes))
        return output


class CustomizingListController(AbstractController):
    """ Customizing list page """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_root(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        cust_reader = ClientCustomizingReader(self.dispatcher)
        entries = cust_reader.get_client_customizing_list(p_client_id=self.authenticated_client_id)

        return render_template(
            "customizing_list.html",
            entries=entries,
            alias=self.dispatcher.ticket.system_alias)

class CustomizingEditController(AbstractController):
    """ Customizing list page """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        node = request.args.get("node", 0, type=str)
        cust_reader = ClientCustomizingReader(self.dispatcher)
        entry = cust_reader.get_client_customizing_entry(self.requested_client_id, node)
        return render_template("customizing_edit.html", client=self.requested_client_id, node=node, entry=entry, alias=self.dispatcher.ticket.system_alias) # pylint: disable=C0301

class CustomizingSaveController(AbstractController):
    """ Customizing save """

    def execute(self):
        """ Builds and returns the page """
        try:
            self._authenticate_minding_requested_client(must_be_admin=True)
        except AuthenticationError as authentication_error:
            return authentication_error.output

        client_db = self.requested_client_database
        node = request.form["node"]
        customizing = request.form["customizing"]
        customizing = customizing.replace("\r\n", "\n") # Windows fix

        if node == _DATABUS_DB_NODE:
            client_db.customizing = customizing
        elif node == _CLIENTS_DB_NODE:
            client_db.client_master_data = customizing
        else:
            external_file = self.dispatcher.external_config_file_manager.get_file(self.requested_client_id, node) # pylint: disable=C0301
            external_file.file_content = customizing

        url = url_for("_customizing_list")
        url += "?" + AbstractController._get_cache_buster()
        return redirect(url, code=302)



""" Module to test issue 116
https://github.com/keremkoseoglu/databus/issues/116
"""
from typing import List
from databus.puller.abstract_exchange import AbstractExchange, ExchangeSettings, \
    ExchangeFolder, ExchangeFolderParent
from databus.passenger.abstract_passenger import AbstractPassenger
from databus.client.log import Log

class Exchange116(AbstractExchange):
    """ Test class """
    def notify_passengers_seated(self, p_seated_passengers: List[AbstractPassenger]):
        """ Not needed for this test """

    @property
    def settings(self) -> ExchangeSettings:
        return ExchangeSettings(
            p_email="user@yourexchange.com",
            p_username="user",
            p_password="pass",
            p_server="your_exchange_server.com")

def test116(action: int):
    """ Runs the test for issue 116 """
    log = Log()
    exchange = Exchange116(p_log=log)
    passengers = exchange.pull()

    if passengers is None or len(passengers) <= 0:
        return

    for passenger in passengers:
        if action == 1:
            exchange.delete_seated_passengers_from_inbox([passenger])
            continue
        if action == 2:
            exchange.forward_seated_passengers([passenger], ["kerem.koseoglu@gmail.com"])
            continue
        if action == 3:
            exchange.trash_seated_passengers([passenger])
            continue
        if action == 4:
            folder = ExchangeFolder(ExchangeFolderParent.INBOX, "kktest1")
            exchange.move_seated_passengers([passenger], folder)
            continue
        if action == 5:
            folder = ExchangeFolder(ExchangeFolderParent.ROOT, "kktest2")
            exchange.move_seated_passengers([passenger], folder)
            continue
